# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import re
from urllib.parse import urljoin
from typing import Dict
import requests

import pyincore.globals as pyglobals

logger = pyglobals.LOGGER

from pyincore import IncoreClient
from pyincore.models.fragilitycurveset import FragilityCurveSet
from pyincore.models.repaircurveset import RepairCurveSet
from pyincore.models.restorationcurveset import RestorationCurveSet
from pyincore.models.mappingset import MappingSet
from pyincore.utils import return_http_response

# add more types if needed
known_types = {
    "java.lang.String": "str",
    "double": "float",
    "int": "int",
    "str": "str"
}

# add more operators if needed
known_operators = {
    "EQ": "==",
    "EQUALS": "==",
    "NEQUALS": "!=",
    "GT": ">",
    "GE": ">=",
    "LT": "<",
    "LE": "<=",
    "NMATCHES": "",
    "MATCHES": ""
}


class MappingSubject(object):
    def __init__(self):
        self.schema = str()
        self.inventory = None  # Feature Collection


class MappingRequest(object):
    def __init__(self):
        self.params = dict()
        self.subject = MappingSubject()


class MappingResponse(object):
    def __init__(self, sets: Dict[str, any]=dict(), mapping: Dict[str, str]=dict()):
        self.sets = sets
        self.mapping = mapping


class Dfr3Service:
    """DFR3 service client.

    Args:
        client (IncoreClient): Service authentication.

    """

    def __init__(self, client: IncoreClient):
        self.client = client
        self.base_mapping_url = urljoin(client.service_url, 'dfr3/api/mappings/')


    def get_dfr3_set(self, dfr3_id: str, timeout=(30, 600), **kwargs):
        """Get specific DFR3 set.

        Args:
            dfr3_id (str): ID of the DFR3 set.
            timeout (tuple): Timeout for the request.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            obj: HTTP response with search results.

        """
        url = urljoin(self.base_dfr3_url, dfr3_id)
        r = self.client.get(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def delete_dfr3_set(self, dfr3_id: str, timeout=(30, 600), **kwargs):
        """Delete specific DFR3 set.
            Args:
                dfr3_id (str): ID of the DFR3 set.
                timeout (tuple): Timeout for the request.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                obj: HTTP response with return results.
        """
        url = urljoin(self.base_dfr3_url, dfr3_id)
        r = self.client.delete(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def batch_get_dfr3_set(self, dfr3_id_lists: list):
        """This method is intended to replace batch_get_dfr3_set in the future. It retrieve dfr3 sets
        from services using id and instantiate DFR3Curveset objects in bulk.

        Args:
            dfr3_id_lists (list): A list of ids.

        Returns:
            list: A list of dfr3curve objects.

        """
        batch_dfr3_sets = {}
        for id in dfr3_id_lists:
            dfr3_set = self.get_dfr3_set(id)
            instance = self.__class__.__name__
            if instance == 'FragilityService':
                batch_dfr3_sets[id] = FragilityCurveSet(dfr3_set)
            elif instance == 'RepairService':
                batch_dfr3_sets[id] = RepairCurveSet(dfr3_set)
            elif instance == 'RestorationService':
                batch_dfr3_sets[id] = RestorationCurveSet(dfr3_set)
            else:
                raise ValueError("Only fragility and repair services are currently supported")

        return batch_dfr3_sets

    def search_dfr3_sets(self, text: str, skip: int = None, limit: int = None, timeout=(30, 600), **kwargs):
        """Search DFR3 sets based on a specific text.

        Args:
            text (str): Text to search by.
            skip (int):  Skip the first n results, default None.
            limit (int): Limit number of results to return, default None.
            timeout (tuple): Timeout for the request.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            obj: HTTP response with search results.

        """
        url = urljoin(self.base_dfr3_url, "search")
        payload = {"text": text}
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def create_dfr3_set(self, dfr3_set: dict, timeout=(30, 600), **kwargs):
        """Create DFR3 set on the server. POST API endpoint call.

        Args:
            dfr3_set (dict): Set of DFR3 jsons.
            timeout (tuple): Timeout for the request.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            obj: HTTP POST Response. Returned value model of the created DFR3 set.

        """
        url = self.base_dfr3_url
        r = self.client.post(url, json=dfr3_set, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def match_inventory(self, mapping: MappingSet, inventories: list, entry_key: str, add_info: list = None):
        """This method is intended to replace the match_inventory method in the future. The functionality is same as
        match_inventory but instead of dfr3_sets in plain json, dfr3 curves will be represented in
        FragilityCurveSet Object.

        Args:
            mapping (obj): MappingSet Object that has the rules and entries.
            inventories (list): A list of inventories. Each item is a casted fiona object
            entry_key (str): keys such as PGA, pgd, and etc.
            add_info (None, dict): additional information that used to match rules, e.g. retrofit strategy per building.

        Returns:
             dict: A dictionary of {"inventory id": FragilityCurveSet object}.

        """
        dfr3_sets = {}

        # loop through inventory to match the rules
        matched_curve_ids = []
        for inventory in inventories:
            if "occ_type" in inventory["properties"] and \
                    inventory["properties"]["occ_type"] is None:
                inventory["properties"]["occ_type"] = ""
            if "efacility" in inventory["properties"] and \
                    inventory["properties"]["efacility"] is None:
                inventory["properties"]["efacility"] = ""

            # if additional information presented, merge inventory properties with that additional information
            if add_info is not None:
                for add_info_row in add_info:
                    if inventory["properties"].get("guid") is not None and \
                            add_info_row.get("guid") is not None and \
                            inventory["properties"].get("guid") == add_info_row.get("guid"):
                        inventory["properties"].update(add_info_row)
                        break  # assume no duplicated guid

            for m in mapping.mappings:
                # for old format rule matching [[]]
                # [[ and ] or [ and ]]
                if isinstance(m.rules, list):
                    if self._property_match_legacy(rules=m.rules, properties=inventory["properties"]):
                        curve = m.entry[entry_key]
                        dfr3_sets[inventory['id']] = curve

                        # if it's string:id; then need to fetch it from remote and cast to fragility3curve object
                        if isinstance(curve, str) and curve not in matched_curve_ids:
                            matched_curve_ids.append(curve)

                        # use the first match
                        break

                # for new format rule matching {"AND/OR":[]}
                # {"AND": [xx, "OR": [yy, yy], "AND": {"OR":["zz", "zz"]]}
                elif isinstance(m.rules, dict):
                    if self._property_match(rules=m.rules, properties=inventory["properties"]):
                        curve = m.entry[entry_key]
                        dfr3_sets[inventory['id']] = curve

                        # if it's string:id; then need to fetch it from remote and cast to fragility3curve object
                        if isinstance(curve, str) and curve not in matched_curve_ids:
                            matched_curve_ids.append(curve)

                        # use the first match
                        break

        batch_dfr3_sets = self.batch_get_dfr3_set(matched_curve_ids)

        # replace the curve id in dfr3_sets to the dfr3 curve
        for inventory_id, curve_item in dfr3_sets.items():
            if isinstance(curve_item, FragilityCurveSet) or isinstance(curve_item, RepairCurveSet) \
                    or isinstance(curve_item, RestorationCurveSet):
                pass
            elif isinstance(curve_item, str):
                dfr3_sets[inventory_id] = batch_dfr3_sets[curve_item]
            else:
                raise ValueError(
                    "Cannot realize dfr3_set entry. The entry has to be either remote id string; or dfr3curve object!")

        return dfr3_sets

    def match_list_of_dicts(self, mapping: MappingSet, inventories: list, entry_key: str):
        """This method is same as match_inventory, except it takes a simple list of dictionaries that contains the items
        to be mapped in the rules. The match_inventory method takes a list of fiona objects

        Args:
            mapping (obj): MappingSet Object that has the rules and entries.
            inventories (list): A list of inventories. Each item of the list is a simple dictionary
            entry_key (str): keys such as PGA, pgd, and etc.

        Returns:
             dict: A dictionary of {"inventory id": FragilityCurveSet object}.

        """
        dfr3_sets = {}

        # loop through inventory to match the rules
        matched_curve_ids = []
        for inventory in inventories:
            for m in mapping.mappings:
                # for old format rule matching [[]]
                if isinstance(m.rules, list):
                    if self._property_match_legacy(rules=m.rules, properties=inventory):
                        curve = m.entry[entry_key]
                        dfr3_sets[inventory['id']] = curve

                        # if it's string:id; then need to fetch it from remote and cast to fragility3curve object
                        if isinstance(curve, str) and curve not in matched_curve_ids:
                            matched_curve_ids.append(curve)

                        # use the first match
                        break

                # for new format rule matching {"AND/OR":[]}
                elif isinstance(m.rules, dict):
                    if self._property_match(rules=m.rules, properties=inventory):
                        curve = m.entry[entry_key]
                        dfr3_sets[inventory['guid']] = curve

                        # if it's string:id; then need to fetch it from remote and cast to fragility3curve object
                        if isinstance(curve, str) and curve not in matched_curve_ids:
                            matched_curve_ids.append(curve)

                        # use the first match
                        break
                    
        batch_dfr3_sets = self.batch_get_dfr3_set(matched_curve_ids)

        # replace the curve id in dfr3_sets to the dfr3 curve
        for inventory_id, curve_item in dfr3_sets.items():
            if isinstance(curve_item, FragilityCurveSet) or isinstance(curve_item, RepairCurveSet) \
                    or isinstance(curve_item, RestorationCurveSet):
                pass
            elif isinstance(curve_item, str):
                dfr3_sets[inventory_id] = batch_dfr3_sets[curve_item]
            else:
                raise ValueError(
                    "Cannot realize dfr3_set entry. The entry has to be either remote id string; or dfr3curve object!")

        return dfr3_sets

    @staticmethod
    def _property_match_legacy(rules, properties):
        """A method to determine whether current set of rules rules applied to the inventory row (legacy rule format).

        Args:
            rules (obj): [[A and B] or [C and D]]
            properties (dict): A dictionary that contains properties of the inventory row.

        Returns:
            True or False

        """

        # if there's no condition, it indicates a match
        if rules == [[]] or rules == [] or rules == [None]:
            return True

        else:
            # rules = [[A and B], OR [C and D], OR [E and F]]
            or_matched = [False for i in range(len(rules))]  # initiate all false state outer list
            for i, and_rules in enumerate(rules):
                and_matched = [False for j in range(len(and_rules))]  # initialte all false state for inner list
                for j, rule in enumerate(and_rules):
                    # evaluate, return True or False. And place it in the corresponding place
                    and_matched[j] = Dfr3Service._eval_criterion(rule, properties)

                # for inner list, AND boolean applied
                if all(and_matched):
                    or_matched[i] = True

        # for outer list, OR boolean is appied
        return any(or_matched)

    @staticmethod
    def _property_match(rules, properties):
        """A method to determine whether current set of rules applied to the inventory row (legacy rule format).

        Args:
            rules (dict):  e.g. {"AND": ["int archetype EQUALS 1", "int retrofit_level EQUALS 0”,
            {“OR": ["int archetype EQUALS 1", "int archetype EQUALS 1"]}]
            properties (dict): dictionary that contains properties of the inventory row

        Returns:
            True or False

        """
        # if without any condition, consider it as a match
        if rules == {}:
            return True
        else:
            boolean = list(rules.keys())[0]  # AND or OR
            criteria = rules[boolean]

            matches = []
            for criterion in criteria:
                # Recursively parse and evaluate the rules with boolean
                if isinstance(criterion, dict):
                    matches.append(Dfr3Service._property_match(criterion, properties))
                # Base case: evaluate the rule and return match=true/false
                elif isinstance(criterion, str):
                    matches.append(Dfr3Service._eval_criterion(criterion, properties))
                else:
                    raise ValueError("Cannot evaluate criterion, unsupported format!")

            if boolean.lower() == "and":
                return all(matches)
            elif boolean.lower() == "or":
                return any(matches)
            else:
                raise ValueError("boolean " + boolean + " not supported!")

    @staticmethod
    def _eval_criterion(rule, properties):
        """A method to evaluate individual rule and see if it appies to a certain inventory row.

        Args:
            rule (str): # e.g. "int no_stories EQ 1",
            properties (dict): dictionary of properties of an invnetory item. e.g. {"guid":xxx, "num_stories":xxx, ...}

        Returns:
            True or False which indicates a match.

        """
        matched = False
        elements = rule.split(" ", 3)
        # the format of a rule is always: rule_type + rule_key + rule_operator + rule_value
        # e.g. "int no_stories EQ 1",
        # e.g. "int year_built GE 1992",
        # e.g. "java.lang.String Soil EQUALS Upland",
        # e.g. "java.lang.String struct_typ EQUALS W2"

        rule_type = elements[0]  # e.g. int, str, double, java.lang.String, etc...
        if rule_type not in known_types.keys():
            raise ValueError(rule_type + " Unknown. Cannot parse the rules of this mapping!")

        rule_key = elements[1]  # e.g. no_storeis, year_built, etc...

        rule_operator = elements[2]  # e.g. EQ, GE, LE, EQUALS
        if rule_operator not in known_operators.keys():
            raise ValueError(rule_operator + " Unknown. Cannot parse the rules of this mapping!")

        rule_value = elements[3].strip('\'').strip('\"')

        if rule_key in properties.keys():
            # validate if the rule is written correctly by comparing variable type, e.g. no_stories properties
            # should be integer
            if isinstance(properties[rule_key], eval(known_types[rule_type])):
                # additional steps to strip "'" for string matches
                if known_types[rule_type] == 'str':
                    if rule_operator == "MATCHES":
                        matched = bool(re.search(rule_value, properties[rule_key]))
                    elif rule_operator == "NMATCHES":
                        matched = not bool(re.search(rule_value, properties[rule_key]))
                    else:
                        matched = eval(
                            '"{0}"'.format(properties[rule_key]) + known_operators[rule_operator] + '"{0}"'.format(
                                rule_value))
                else:
                    matched = eval(str(properties[rule_key]) + known_operators[rule_operator] + rule_value)
            else:
                raise ValueError("Mismatched datatype found in the mapping rule: " + rule +
                                 ". Datatype found in the dataset for " + rule_key + " : "
                                 + str(type(properties[rule_key])) + ". Please review the mapping being used.")

        return matched

    @staticmethod
    def extract_inventory_class_legacy(rules):
        """This method will extract the inventory class name from a mapping rule. E.g. PWT2/PPP1

        Args:
            rules (list): The outer list is applying "OR" rule and the inner list is applying an "AND" rule. e.g. list(["java.lang.String utilfcltyc EQUALS 'PWT2'"],["java.lang.String utilfcltyc EQUALS 'PPP1'"])

        Returns:
            inventory_class (str): extracted inventory class name. "/" stands for or and "+" stands for and

        """
        if rules == [[]] or rules == [] or rules == [None]:
            return "NA"
        else:
            inventory_class = ""
            for i, and_rules in enumerate(rules):
                if i != 0:
                    inventory_class += "/"

                for j, rule in enumerate(and_rules):
                    if j != 0:
                        inventory_class += "+"
                    inventory_class += rule.split(" ")[3].strip('\'').strip('\"')
            return inventory_class

    @staticmethod
    def extract_inventory_class(rules):
        """This method will extract the inventory class name from a mapping rule. E.g. PWT2/PPP1

        Args:
            rules (dict): e.g. { "AND": ["java.lang.String utilfcltyc EQUALS 'PWT2'", "java.lang.String utilfcltyc EQUALS 'PPP1'"]}

        Returns:
            inventory_class (str): extracted inventory class name. "/" stands for or and "+" stands for and

        """
        if rules == {}:
            return "NA"
        else:
            inventory_class = []
            boolean = list(rules.keys())[0]  # AND or OR
            criteria = rules[boolean]
            for criterion in criteria:
                if isinstance(criterion, dict):
                    inventory_class.append(Dfr3Service.extract_inventory_class(criterion))
                elif isinstance(criterion, str):
                    inventory_class.append(criterion.split(" ")[3].strip('\'').strip('\"'))
                else:
                    raise ValueError("Cannot evaluate criterion, unsupported format!")

            if boolean.lower() == "and":
                return "+".join(inventory_class)
            elif boolean.lower() == "or":
                return "/".join(inventory_class)
            else:
                raise ValueError("boolean " + boolean + " not supported!")

    def create_mapping(self, mapping_set: dict, timeout=(30, 600), **kwargs):
        """Create DFR3 mapping on the server. POST API endpoint call.

        Args:
            mapping_set (dict): Mapping set, relationship between inventories (buildings, bridges etc.)
                and DFR3 sets.
            timeout (tuple): Timeout for the request.
            **kwargs: Additional keyword arguments.

        Returns:
            obj: HTTP POST Response. Returned value model of the created mapping set.

        """
        url = self.base_mapping_url
        r = self.client.post(url, json=mapping_set, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_mappings(self, hazard_type: str = None, inventory_type: str = None, mapping_type: str = None,
                     creator: str = None, space: str = None, skip: int = None, limit: int = None,
                     timeout=(30, 600), **kwargs):
        """Get the set of mappings. Mapping is a relationship between inventories (buildings, bridges
            etc.) and DFR3 sets.

        Args:
            hazard_type (str): Hazard type filter, default None.
            inventory_type (str): Inventory type, default None.
            mapping_type (str): mapping type, default None.
            creator (str): creator’s username, default None.
            space (str): Name of space, default None.
            skip (int):  Skip the first n results, default None.
            limit (int): Limit number of results to return, default None.
            timeout (tuple): Timeout for the request, default (30, 600).
            **kwargs: Additional keyword arguments.

        Returns:
            obj: HTTP response with search results.

        """
        url = self.base_mapping_url
        payload = {}

        if hazard_type is not None:
            payload['hazard'] = hazard_type
        if inventory_type is not None:
            payload['inventory'] = inventory_type
        if mapping_type is not None:
            payload['mappingType'] = mapping_type
        if creator is not None:
            payload['creator'] = creator
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit
        if space is not None:
            payload['space'] = space

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_mapping(self, mapping_id, timeout=(30, 600), **kwargs):
        """Get specific inventory mapping.

        Args:
            mapping_id (str): ID of the Mapping set.
            timeout (tuple): Timeout for the request, default (30, 600).
            **kwargs: Additional keyword arguments.

        Returns:
            obj: HTTP response with search results.

        """
        url = urljoin(self.base_mapping_url, mapping_id)
        r = self.client.get(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def delete_mapping(self, mapping_id, timeout=(30, 600), **kwargs):
        """delete specific inventory mappings.

        Args:
            mapping_id (str): ID of the Mapping set.
            timeout (tuple): Timeout for the request, default (30, 600).
            **kwargs: Additional keyword arguments.

        Returns:
            obj: HTTP response with return results.

        """
        url = urljoin(self.base_mapping_url, mapping_id)
        r = self.client.delete(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()



# Copyright (c) 2021 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
import json


class HHRSOutputProcess:
    """This class converts results outputs of housing household recovery sequential analysis to certain format."""

    @staticmethod
    def get_hhrs_stage_count(timesteps, hhrs_df, filename_json="hhrs_stage_count.json"):
        """

        Args:
            timesteps: timesteps in the unit of month ["0", "6", "12", "24", "48", etc]
            hhrs_df: pandas dataframe of the output of housingrecoverysequential
            filename_json: the name of the json file to store the output

        Returns:

        """

        hhrs_stage_count = {}
        for t in timesteps:
            stage = hhrs_df[t]
            hhrs_stage_count[t] = [int((stage == 1.0).sum()), int((stage == 2.0).sum()), int((stage == 3.0).sum()),
                                   int((stage == 4.0).sum()), int((stage == 5.0).sum())]

        if filename_json:
            with open(filename_json, "w") as outfile:
                json.dump(hhrs_stage_count, outfile, indent=2)

        return hhrs_stage_count




# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import csv
import glob
import json
import os

import fiona
import numpy
import pandas as pd
import geopandas as gpd
import rasterio
import warnings
from pyincore import DataService

warnings.filterwarnings("ignore", "", UserWarning)


class Dataset:
    """Dataset.

    Args:
        metadata (dict): Dataset metadata.

    """

    def __init__(self, metadata):
        self.metadata = metadata

        # For convenience instead of having to dig through the metadata for these
        self.data_type = metadata["dataType"]
        self.format = metadata["format"]
        self.id = metadata["id"]
        self.file_descriptors = metadata["fileDescriptors"]
        self.local_file_path = None

        self.readers = {}

    @classmethod
    def from_data_service(cls, id: str, data_service: DataService):
        """Get Dataset from Data service, get metadata as well.

        Args:
            id (str): ID of the Dataset.
            data_service (obj): Data service.

        Returns:
            obj: Dataset from Data service.

        """
        metadata = data_service.get_dataset_metadata(id)
        instance = cls(metadata)
        instance.cache_files(data_service)
        return instance

    @classmethod
    def from_json_str(cls, json_str, data_service: DataService = None, file_path=None):
        """Get Dataset from json string.

        Args:
            json_str (str): JSON of the Dataset.
            data_service (obj): Data Service class.
            file_path (str): File path.

        Returns:
            obj: Dataset from JSON.

        """
        instance = cls(json.loads(json_str))

        # download file and set local file path using metadata from services
        if data_service is not None:
            instance.cache_files(data_service)

        # if there is local files associates with the dataset
        elif file_path is not None:
            instance.local_file_path = file_path

        else:
            raise ValueError("You have to either use data services, or given pass local file path.")

        return instance

    @classmethod
    def from_file(cls, file_path, data_type):
        """Get Dataset from the file.

        Args:
            file_path (str): File path.
            data_type (str): Data type.

        Returns:
            obj: Dataset from file.

        """
        metadata = {"dataType": data_type,
                    "format": '',
                    "fileDescriptors": [],
                    "id": file_path}
        instance = cls(metadata)
        instance.local_file_path = file_path
        return instance

    @classmethod
    def from_dataframe(cls, dataframe, name, data_type, index=False):
        """Get Dataset from Panda's DataFrame.

        Args:
            dataframe (obj): Panda's DataFrame.
            name (str): filename.
            data_type (str): Incore data type, e.g. incore:xxxx or ergo:xxxx
            index (bool): Store the index column

        Returns:
            obj: Dataset from file.

        """
        dataframe.to_csv(name, index=index)
        return Dataset.from_file(name, data_type)

    @classmethod
    def from_csv_data(cls, result_data, name, data_type):
        """Get Dataset from CSV data.

        Args:
            result_data (obj): Result data and metadata.
            name (str): A CSV filename.
            data_type (str): Incore data type, e.g. incore:xxxx or ergo:xxxx

        Returns:
            obj: Dataset from file.

        """
        if len(result_data) > 0:
            with open(name, 'w') as csv_file:
                # Write the parent ID at the top of the result data, if it is given
                writer = csv.DictWriter(csv_file, dialect="unix", fieldnames=result_data[0].keys())
                writer.writeheader()
                writer.writerows(result_data)
        return Dataset.from_file(name, data_type)

    @classmethod
    def from_json_data(cls, result_data, name, data_type):
        """Get Dataset from JSON data.

        Args:
            result_data (obj): Result data and metadata.
            name (str): A JSON filename.
            data_type (str): Incore data type, e.g. incore:xxxx or ergo:xxxx

        Returns:
            obj: Dataset from file.

        """
        if len(result_data) > 0:
            with open(name, 'w') as json_file:
                json_dumps_str = json.dumps(result_data, indent=4)
                json_file.write(json_dumps_str)
        return Dataset.from_file(name, data_type)

    def cache_files(self, data_service: DataService):
        """Get the set of fragility data, curves.

        Args:
            data_service (obj): Data service.

        Returns:
            str: A path to the local file.

        """
        if self.local_file_path is not None:
            return
        self.local_file_path = data_service.get_dataset_blob(self.id)
        return self.local_file_path

    """Utility methods for reading different standard file formats"""

    def get_inventory_reader(self):
        """Utility method for reading different standard file formats: Set of inventory.

        Returns:
            obj: A Fiona object.

        """
        filename = self.local_file_path
        if os.path.isdir(filename):
            layers = fiona.listlayers(filename)
            if len(layers) > 0:
                # for now, open a first shapefile
                return fiona.open(filename, layer=layers[0])
        else:
            return fiona.open(filename)

    def get_json_reader(self):
        """Utility method for reading different standard file formats: json reader.

        Returns:
            obj: A json model data.

        """
        if "json" not in self.readers:
            filename = self.local_file_path
            if os.path.isdir(filename):
                files = glob.glob(filename + "/*.json")
                if len(files) > 0:
                    filename = files[0]

            with open(filename, 'r') as f:
                return json.load(f)

        return self.readers["json"]

    def get_raster_value(self, x, y):
        """Utility method for reading different standard file formats: raster value.

        Args:
            x (float): X coordinate.
            y (float): Y coordinate.
        Returns:
            numpy.array: Hazard values.

        """
        if "raster" not in self.readers:
            filename = self.local_file_path
            if os.path.isdir(filename):
                files = glob.glob(filename + "/*.tif")
                if len(files) > 0:
                    filename = files[0]
            self.readers["raster"] = rasterio.open(filename)

        hazard = self.readers["raster"]
        row, col = hazard.index(x, y)
        # assume that there is only 1 band
        data = hazard.read(1)
        xmin, ymin, xmax, ymax = hazard.bounds
        if x < xmin or x > xmax or y < ymin or y > ymax:
            return None
        # TODO check threshold
        return float(data[row, col])

    def get_csv_reader(self):
        """Utility method for reading different standard file formats: csv reader.

         Returns:
             obj: CSV reader.

         """
        if "csv" not in self.readers:
            filename = self.local_file_path
            if os.path.isdir(filename):
                files = glob.glob(filename + "/*.csv")
                if len(files) > 0:
                    filename = files[0]

            csvfile = open(filename, 'r')
            return csv.DictReader(csvfile)

        return self.readers["csv"]

    def get_csv_reader_std(self):
        """Utility method for reading different standard file formats: csv reader.

         Returns:
             obj: CSV reader.

         """
        if "csv" not in self.readers:
            filename = self.local_file_path
            if os.path.isdir(filename):
                files = glob.glob(filename + "/*.csv")
                if len(files) > 0:
                    filename = files[0]

            csvfile = open(filename, 'r')
            return csv.reader(csvfile)

        return self.readers["csv"]

    def get_file_path(self, type='csv'):
        """Utility method for reading different standard file formats: file path.

        Args:
            type (str): A file type.

        Returns:
            str: File name and path.

        """
        filename = self.local_file_path
        if os.path.isdir(filename):
            files = glob.glob(filename + "/*." + type)
            if len(files) > 0:
                filename = files[0]

        return filename

    def get_dataframe_from_csv(self, low_memory=True, delimiter=None):
        """Utility method for reading different standard file formats: Pandas DataFrame from csv.

        Args:
            low_memory (bool): A flag to suppress warning. Pandas is guessing dtypes for each column
            if column dtype is not specified which is very memory demanding.

        Returns:
            obj: Panda's DataFrame.

        """
        filename = self.get_file_path('csv')
        df = pd.DataFrame()
        if os.path.isfile(filename):
            df = pd.read_csv(filename, header="infer", low_memory=low_memory, delimiter=delimiter)
        return df

    def get_dataframe_from_shapefile(self):
        """Utility method for reading different standard file formats: GeoDataFrame from shapefile.

        Returns:
            obj: Geopanda's GeoDataFrame.

        """
        # read shapefile directly by Geopandas.read_file()
        # It will preserve CRS information also
        gdf = gpd.read_file(self.local_file_path)

        return gdf

    def close(self):
        for key in self.readers:
            self.readers[key].close()

    def __del__(self):
        self.close()


class DamageRatioDataset:
    """For backwards compatibility until analyses are updated.

    Args:
        filename (str): CSV file with damage ratios.

    """

    def __init__(self, filename):
        self.damage_ratio = None
        csvfile = open(filename, 'r')
        reader = csv.DictReader(csvfile)
        self.damage_ratio = []
        for row in reader:
            self.damage_ratio.append(row)


class InventoryDataset:
    """For backwards compatibility until analyses are updated.

    Args:
        filename (str): file with GIS layers.

    """

    def __init__(self, filename):
        self.inventory_set = None
        if os.path.isdir(filename):
            layers = fiona.listlayers(filename)
            if len(layers) > 0:
                # for now, open a first shapefile
                self.inventory_set = fiona.open(filename, layer=layers[0])
        else:
            self.inventory_set = fiona.open(filename)

    def close(self):
        self.inventory_set.close()

    def __del__(self):
        self.close()



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import numpy as np
import pandas as pd
import time

import warnings

from pyincore import BaseAnalysis, RepairService
from pyincore.analyses.buildingdamage.buildingutil import BuildingUtil


class ResidentialBuildingRecovery(BaseAnalysis):
    """
    This analysis computes the recovery time needed for each residential building from any damage states to receive the
    full restoration. Currently, supported hazards are tornadoes.

    The methodology incorporates the multi-layer Monte Carlo simulation approach and determines the two-step recovery
    time that includes delay and repair. The delay model was modified based on the REDi framework and calculated the
    end-result outcomes resulted from delay impeding factors such as post-disaster inspection, insurance claim, and
    government permit. The repair model followed the FEMA P-58 approach and was controlled by fragility functions.

    The outputs of this analysis is a CSV file with time-stepping recovery probabilities at the building level.

    Contributors
        | Science: Wanting Lisa Wang, John W. van de Lindt
        | Implementation: Wanting Lisa Wang, Gowtham Naraharisetty, and NCSA IN-CORE Dev Team

    Related publications
        Wang, Wanting Lisa, and John W. van de Lindt. "Quantitative Modeling of Residential Building Disaster Recovery
        and Effects of Pre-and Post-event Policies." International Journal of Disaster Risk Reduction (2021): 102259.

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        self.repairsvc = RepairService(incore_client)

        super(ResidentialBuildingRecovery, self).__init__(incore_client)

    def run(self):
        """Executes the residential building recovery analysis.

        Returns:
            bool: True if successful, False otherwise.

        """
        # TODO: Start using seed
        seed = self.get_parameter("seed")

        num_samples = self.get_parameter("num_samples")
        result_name = self.get_parameter("result_name")

        repair_key = self.get_parameter("repair_key")
        if repair_key is None:
            repair_key = BuildingUtil.DEFAULT_REPAIR_KEY
            self.set_parameter("repair_key", repair_key)

        buildings = self.get_input_dataset("buildings").get_inventory_reader()
        buildings = list(buildings)
        sample_damage_states = self.get_input_dataset("sample_damage_states").get_dataframe_from_csv(low_memory=False)
        socio_demographic_data = self.get_input_dataset("socio_demographic_data").get_dataframe_from_csv(
            low_memory=False)
        # socio_demographic_data.set_index("guid", inplace=True)
        financial_resources = self.get_input_dataset("financial_resources").get_dataframe_from_csv(low_memory=False)
        redi_delay_factors = self.get_input_dataset("delay_factors").get_dataframe_from_csv(low_memory=False)

        # Returns dataframe
        recovery_results = self.residential_recovery(buildings, sample_damage_states, socio_demographic_data,
                                                     financial_resources, redi_delay_factors, num_samples)
        self.set_result_csv_data("residential_building_recovery", recovery_results, result_name, "dataframe")

        return True

    def residential_recovery(self, buildings, sample_damage_states, socio_demographic_data, financial_resources,
                             redi_delay_factors, num_samples):
        """
        Calculates residential building recovery for buildings

        Args:
            buildings(list): Buildings dataset
            sample_damage_states (pd.DataFrame): Sample damage states
            socio_demographic_data (pd.DataFrame): Socio-demographic data for household income groups
            financial_resources (pd.DataFrame): Financial resources by household income groups
            redi_delay_factors (pd.DataFrame): Delay factors based on REDi framework
            num_samples (int): number of sample scenarios to use

        Returns:
            dict: dictionary with id/guid and residential recovery for each quarter

        """

        start_household_income_prediction = time.process_time()
        household_income_prediction = ResidentialBuildingRecovery.household_income_prediction(socio_demographic_data,
                                                                                              num_samples)
        end_start_household_income_prediction = time.process_time()
        print("Finished executing household_income_prediction() in " +
              str(end_start_household_income_prediction - start_household_income_prediction) + " secs")

        household_aggregation = ResidentialBuildingRecovery.household_aggregation(household_income_prediction)
        end_household_aggregation = time.process_time()
        print("Finished executing household_aggregation() in " +
              str(end_household_aggregation - end_start_household_income_prediction) + " secs")

        financing_delay = ResidentialBuildingRecovery.financing_delay(household_aggregation, financial_resources)
        end_financing_delay = time.process_time()
        print("Finished executing financing_delay() in " +
              str(end_financing_delay - end_household_aggregation) + " secs")

        total_delay = ResidentialBuildingRecovery.total_delay(sample_damage_states, redi_delay_factors,
                                                                financing_delay)
        end_total_delay = time.process_time()
        print("Finished executing total_delay() in " + str(end_total_delay - end_financing_delay) + " secs")

        recovery = self.recovery_rate(buildings, sample_damage_states, total_delay)
        end_recovery = time.process_time()
        print("Finished executing recovery_rate() in " + str(end_recovery - end_total_delay) + " secs")

        time_stepping_recovery = ResidentialBuildingRecovery.time_stepping_recovery(recovery)
        end_time_stepping_recovery = time.process_time()
        print("Finished executing time_stepping_recovery() in " +
              str(end_time_stepping_recovery - end_recovery) + " secs")

        result = time_stepping_recovery

        end_time = time.process_time()
        print("Analysis completed in " + str(end_time - start_household_income_prediction) + " secs")

        return result

    @staticmethod
    def household_income_prediction(income_groups, num_samples):
        """ Get Income group prediction for each household

        Args:
            income_groups (pd.DataFrame): Socio-demographic data with household income group prediction.
            num_samples (int): Number of sample scenarios.

        Returns:
            pd.DataFrame: Income group prediction for each household

        """

        blockid = income_groups.groupby('blockid')
        prediction_results = pd.DataFrame()

        for name, group in blockid:
            # Prepare data for numpy processing
            group_size = group.shape[0]
            group_hhinc_values = group['hhinc'].values

            # Compute normal distribution parameters from group data

            # suppress nan warnings.
            # 1. Mean of nans - https://stackoverflow.com/questions/29688168/
            # mean-nanmean-and-warning-mean-of-empty-slice
            # 2. Mean of empty slice - https://stackoverflow.com/questions/43301112/
            # mean-of-empty-slice-and-degrees-of-freedom-0
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                mean = np.nanmean(group_hhinc_values)
                std = np.nanstd(group_hhinc_values)

            if np.isnan(mean):
                mean = 3
            if np.isnan(std):
                std = 0
            # Directly compute the indices of NaN values in the hhinc vector
            group_nan_idx = np.where(np.isnan(group_hhinc_values))
            number_nan = len(group_nan_idx[0])

            # Now, generate a numpy matrix to hold the samples for the group
            group_samples = np.zeros((num_samples, group_size))

            for i in range(num_samples):
                # Note to Lisa that this is not the appropriate distribution. Since this case is discrete,
                # the natural choice is a Bernoulli distribution parameterized as to approximate the
                # corresponding normal.
                sample = np.random.normal(mean, std, number_nan)
                sample[np.where(sample > 5)] = 5
                sample[np.where(sample < 1)] = 1
                group_samples[i, :] = group_hhinc_values
                group_samples[i, group_nan_idx[0]] = np.around(sample)
                # Now reassemble into Pandas DataFrame
                group['sample_{}'.format(i)] = group_samples[i, :]

            prediction_results = pd.concat([prediction_results, group], ignore_index=True)

        return prediction_results

    @staticmethod
    def household_aggregation(household_income_predictions):
        """ Gets household aggregation of income groups at the building level.

        Args:
            household_income_predictions (pd.DataFrame): Income group prediction for each household

        Returns:
            pd.DataFrame: Results of household aggregation of income groups at the building level.
        """

        # Drop all unnecessary columns first
        household_income_predictions_dropped = household_income_predictions.drop(columns=['huid', 'blockid', 'hhinc'])
        guid_group = household_income_predictions_dropped.groupby('guid')

        # Obtain sample column names
        colnames = list(household_income_predictions_dropped.columns[1:])

        # Iterate over all groups
        new_groups = []

        for name, group in guid_group:
            # Obtain guids
            local_guids = group['guid']

            # Remove guids
            no_guids = group.drop(columns=['guid']).to_numpy()

            # Compute the maxima of all columns
            maxima = list(no_guids.max(axis=0))

            # Generate a matrix to store the results efficiently
            num_guids = no_guids.shape[0]
            num_cols = len(colnames)
            no_guids_maxima = np.zeros((num_guids, num_cols))

            for i in range(0, num_cols):
                no_guids_maxima[:, i] = maxima[i]

            group_new = pd.DataFrame(no_guids_maxima, columns=colnames, index=group.index)
            group_new.insert(0, 'guid', local_guids)
            new_groups.append(group_new.head(1))

        # Construct a new DataFrame
        household_aggregation_results = pd.concat(new_groups).reset_index(drop=True)

        return household_aggregation_results

    @staticmethod
    def financing_delay(household_aggregated_income_groups, financial_resources):
        """ Gets financing delay, the percentages calculated are the probabilities of housing units financed by
        different resources.

        Args:
            household_aggregated_income_groups (pd.DataFrame): Household aggregation of income groups at the building
                level.
            financial_resources (pd.DataFrame): Financial resources by household income groups.

        Returns:
            pd.DataFrame: Results of financial delay
        """
        colnames = list(household_aggregated_income_groups.columns)[1:]

        # Save guid's for later
        household_guids = household_aggregated_income_groups['guid']

        # Convert household aggregated income to numpy
        samples_np = household_aggregated_income_groups.drop(columns=['guid']).to_numpy()

        # Number of guids
        num_households = household_guids.shape[0]
        num_samples = len(colnames)

        # Convert the sample matrix to numpy
        resources_np = financial_resources.to_numpy()

        # Also, convert financial resources to numpy to perform linear algebra
        hhinc = resources_np[0:5, 0]
        sources = resources_np[:, 1:]

        # Give names to mean and sigma indices
        mean_idx = 5
        sigma_idx = 6

        for household in range(0, num_households):
            for sample in range(0, num_samples):
                value = samples_np[household, sample]
                idx = np.where(hhinc == value)

                # 1. Sample the lognormal distribution vectorially
                lognormal_vec = np.random.lognormal(np.log(sources[mean_idx, :]), sources[sigma_idx, :])

                # 2. Compute the delay using the dot product of the prior vector and sources for the current index,
                # round to one significant figure
                samples_np[household, sample] = np.round(np.dot(lognormal_vec, sources[idx, :].flatten()), 1)

        financing_delay = pd.DataFrame(samples_np, columns=colnames, index=household_aggregated_income_groups.index)
        financing_delay.insert(0, 'guid', household_guids)

        return financing_delay

    @staticmethod
    def total_delay(sample_damage_states, redi_delay_factors, financing_delay):
        """ Calculates total delay by combining financial delay and other factors from REDi framework

        Args:
            sample_damage_states (pd.DataFrame): Building inventory damage states.
            redi_delay_factors (pd.DataFrame): Delay impeding factors such as post-disaster inspection, insurance claim,
                and government permit based on building's damage state.
            financing_delay (pd.DataFrame): Financing delay, the percentages calculated are the probabilities of housing
                units financed by different resources.

        Returns:
            pd.DataFrame: Total delay time of financial delay and other factors from REDi framework.
        """

        # Obtain the column names
        colnames = list(financing_delay.columns)[1:]

        # Perform an inner join to ensure only households with damage states are processed
        merged_delay = pd.merge(financing_delay, sample_damage_states, on='guid')

        # Obtain the guids
        merged_delay_guids = merged_delay['guid']

        # Obtain the damage states
        merged_delay_damage_states = merged_delay['sample_damage_states']

        # Convert to numpy
        samples_np = merged_delay.drop(columns=['guid', 'sample_damage_states']).to_numpy()
        num_samples = len(colnames)

        # First, we decompose redi_delay_factors into two dictionaries that can be used to compute vector operations
        redi_idx = dict(zip(redi_delay_factors['Building_specific_conditions'], redi_delay_factors.index))

        # Next, we produce two intermediate numpy matrices: one for med and one for sdv
        redi_med = redi_delay_factors[['Ins_med', 'Enmo_med', 'Como_med', 'Per_med']].to_numpy()
        redi_sdv = redi_delay_factors[['Ins_sdv', 'Enmo_sdv', 'Como_sdv', 'Per_sdv']].to_numpy()

        # Define indices to facilitate interpretation of the code
        inspection_idx = 0
        engineer_idx = 1
        contractor_idx = 2
        permit_idx = 3

        for i, ds_list in enumerate(list(merged_delay_damage_states)):
            samples_mcs = ds_list.split(",")

            for j in range(num_samples):
                # TODO: ask why there are many more damage states than samples
                # Obtain the index for the corresponding damage state
                dmg_state_idx = redi_idx[samples_mcs[j]]

                # Use the index to select the appropriate mean and stdev vectors
                mean_vec = redi_med[dmg_state_idx, :]
                sdv_vec = redi_sdv[dmg_state_idx, :]

                # Compute the delay vector
                delay_vec = np.random.lognormal(np.log(mean_vec), sdv_vec)

                # Compute the delay using that vector and financing delays, already computed in the prior step
                samples_np[i, j] = np.round(delay_vec[inspection_idx] +
                                            np.max([
                                                delay_vec[engineer_idx],
                                                samples_np[i, j],
                                                delay_vec[contractor_idx]
                                            ]) +
                                            delay_vec[permit_idx])

        total_delay = pd.DataFrame(samples_np, columns=colnames)
        total_delay.insert(0, 'guid', merged_delay_guids)

        return total_delay

    def recovery_rate(self, buildings, sample_damage_states, total_delay):
        """ Gets total time required for each building to receive full restoration. Determined by the combination of
        delay time and repair time

        Args:
            buildings (list): List of buildings
            sample_damage_states (pd.DataFrame): Samples' damage states
            total_delay (pd.DataFrame): Total delay time of financial delay and other factors from REDi framework.

        Returns:
            pd.DataFrame: Recovery rates of all buildings for each sample
        """

        repair_key = self.get_parameter("repair_key")
        repair_sets = self.repairsvc.match_inventory(self.get_input_dataset("dfr3_mapping_set"), buildings, repair_key)
        repair_sets_by_guid = {}  # get repair sets by guid so they can be mapped with output of monte carlo

        # This is sort of a workaround until we define Repair Curve models and abstract this out there
        for i, b in enumerate(buildings):
            repair_sets_by_guid[b["properties"]['guid']] = repair_sets[str(i)]

        # Obtain the column names
        colnames = list(total_delay.columns)[1:]

        # Perform an inner join to ensure only households with damage states are processed
        merged_delay = pd.merge(total_delay, sample_damage_states, on='guid')

        # Obtain the guids
        merged_delay_guids = merged_delay['guid']

        # Obtain the damage states
        merged_delay_damage_states = merged_delay['sample_damage_states']

        # Convert to numpy
        samples_np = merged_delay.drop(columns=['guid', 'sample_damage_states']).to_numpy()
        num_samples = len(colnames)
        num_households = samples_np.shape[0]

        # Generate a long numpy matrix for combined N1, N2 samples
        samples_n1_n2 = np.zeros((num_households, num_samples*num_samples))

        # Now, we define an internal function to take care of the index for the prior case
        # Now, we define an internal function to take care of the index for the prior case
        def idx(x, y):
            return x*num_samples + y

        for household in range(0, num_households):
            # Obtain the damage states
            mapped_repair = repair_sets_by_guid[merged_delay_guids.iloc[household]]
            samples_mcs = merged_delay_damage_states.iloc[household].split(",")

            # Use a lambda to obtain the damage state in numeric form. Note that since damage states are single digits,
            # it suffices to look at the last character and convert into an integer value. Do this computation once
            # per household only.
            extract_ds = lambda x: int(x[-1])
            samples_mcs_ds = list(map(extract_ds, samples_mcs))

            # Now, perform the two nested loops, using the indexing function to simplify the syntax.
            for i in range(0, num_samples):
                state = samples_mcs_ds[i]
                percent_func = np.random.random(num_samples)
                # NOTE: Even though the kwarg name is "repair_time", it actually takes  percent of functionality. DFR3
                # system currently doesn't have a way to represent the name correctly when calculating the inverse.
                repair_time = mapped_repair.repair_curves[state].solve_curve_for_inverse(
                    hazard_values={}, curve_parameters=mapped_repair.curve_parameters, **{"repair_time": percent_func}
                                                                                        ) / 7

                for j in range(0, num_samples):
                    samples_n1_n2[household, idx(i, j)] = round(samples_np[household, i] + repair_time[j], 1)

        # Now, generate all the labels using list comprehension outside the loops
        colnames = [f'sample_{i}_{j}' for i in range(0, num_samples) for j in range(0, num_samples)]
        recovery_time = pd.DataFrame(samples_n1_n2, columns=colnames)
        recovery_time.insert(0, 'guid', merged_delay_guids)

        return recovery_time

    @staticmethod
    def time_stepping_recovery(recovery_results):
        """ Converts results to a time frame. Currently gives results for 16 quarters over 4 year.

        Args:
            recovery_results (pd.DataFrame): Total recovery time of financial delay and other factors from REDi framework.

        Returns:
            pd.DataFrame: Time formatted recovery results.

        """
        time_step = 90 / 7  # a quarter in weeks
        year = 4

        total_time = time_step * np.linspace(0, 4 * year, num=17, endpoint=True)

        # Save guid's for later
        recovery_results_guids = recovery_results['guid']

        # Convert household aggregated income to numpy
        samples_n1_n2 = recovery_results.drop(columns=['guid']).to_numpy()

        # Number of guids
        num_households = recovery_results.shape[0]
        num_samples = samples_n1_n2.shape[1]
        num_times = len(total_time)

        # Generate a numpy to hold the results as desired
        times_np = np.full((num_households, num_times), 1111.0)

        for household in range(0, num_households):
            for i in range(len(total_time)):
                fun_state = np.count_nonzero(samples_n1_n2[household, :] < total_time[i]) / num_samples
                times_np[household, i] = np.round(fun_state, 2)

        colnames = [f'quarter_{i}' for i in range(0, num_times)]

        time_stepping_recovery = pd.DataFrame(times_np, columns=colnames)
        time_stepping_recovery.insert(0, 'guid', recovery_results_guids)

        return time_stepping_recovery

    def get_spec(self):
        """Get specifications of the residential building recovery analysis.

        Returns:
            obj: A JSON object of specifications of the residential building recovery analysis.

        """
        return {
            'name': 'residential-building-recovery',
            'description': 'calculate residential building recovery',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'name of the result',
                    'type': str
                },
                {
                    'id': 'num_samples',
                    'required': True,
                    'description': 'Number of sample scenarios',
                    'type': int
                },
                {
                    'id': 'repair_key',
                    'required': False,
                    'description': 'Repair key to use in mapping dataset',
                    'type': str
                },
                {
                    'id': 'seed',
                    'required': False,
                    'description': 'Initial seed for the probabilistic model',
                    'type': int
                }
            ],
            'input_datasets': [
                {
                    'id': 'buildings',
                    'required': True,
                    'description': 'Building Inventory',
                    'type': ['ergo:buildingInventoryVer4', 'ergo:buildingInventoryVer5', 'ergo:buildingInventoryVer6',
                             'ergo:buildingInventoryVer7']
                },
                {
                    'id': 'dfr3_mapping_set',
                    'required': True,
                    'description': 'DFR3 Mapping Set Object',
                    'type': ['incore:dfr3MappingSet'],
                },
                {
                    'id': 'sample_damage_states',
                    'required': True,
                    'description': 'Sample damage states',
                    'type': ['incore:sampleDamageState']
                },
                {
                    'id': 'socio_demographic_data',
                    'required': True,
                    'description': 'Socio-demographic data with household income group predictions',
                    'type': ['incore:socioDemograhicData']
                },
                {
                    'id': 'financial_resources',
                    'required': True,
                    'description': 'Financial resources by household income groups',
                    'type': ['incore:householdFinancialResources']
                },
                {
                    'id': 'delay_factors',
                    'required': True,
                    'description': 'Delay impeding factors such as post-disaster inspection, insurance claim, '
                                   'and government permit based on building\'s damage state. Provided by REDi framework',
                    'type': ['incore:buildingRecoveryFactors']
                }
            ],
            'output_datasets': [
                {
                    'id': 'residential_building_recovery',
                    'description': 'CSV file of residential building recovery percent',
                    'type': 'incore:buildingRecovery'
                }
            ]
        }



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import copy
import os
import sys
import time

import pandas as pd
import pyomo.environ as pyo
from pyomo.opt import SolverFactory, TerminationCondition
from pyomo.util.infeasible import log_infeasible_constraints

from pyincore import BaseAnalysis, NetworkDataset
from pyincore.analyses.indp.dislocationutils import DislocationUtil
from pyincore.analyses.indp.indpresults import INDPResults
from pyincore.analyses.indp.indputil import INDPUtil
from pyincore.analyses.indp.infrastructureutil import InfrastructureUtil

from pyincore import globals as pyglobals


class INDP(BaseAnalysis):
    """
        This class runs INDP or td-INDP for a given number of time steps and input parameters.This analysis takes a
        decentralized approach to solve the Interdependent Network Design Problem (INDP), a family of
        centralized Mixed-Integer Programming (MIP) models, which find the optimal restoration strategy of disrupted
        networked systems subject to budget and operational constraints.

        Contributors
            | Science: Hesam Talebiyan
            | Implementation: Chen Wang and NCSA IN-CORE Dev Team

        Args:
            incore_client (IncoreClient): Service authentication.
        """

    def __init__(self, incore_client):
        super(INDP, self).__init__(incore_client)

    def run(self):
        # input parameters
        network_type = self.get_parameter("network_type")
        sample_range = self.get_parameter("sample_range")
        MAGS = self.get_parameter("MAGS")
        filter_sce = None
        fail_sce_param = {
            'TYPE': network_type,
            'SAMPLE_RANGE': sample_range,
            'MAGS': MAGS,
            'FILTER_SCE': filter_sce,
        }

        RC = self.get_parameter('RC')
        layers = self.get_parameter('layers')
        method = self.get_parameter('method')

        t_steps = self.get_parameter('t_steps')
        if t_steps is None:
            t_steps = 10

        dynamic_params = None
        dislocation_data_type = self.get_parameter("dislocation_data_type")
        return_model = self.get_parameter("return_model")
        testbed_name = self.get_parameter("testbed_name")

        # more advanced; can hide from user for the first implementation
        # this is for dynamically updating the demands
        bldgs2elec_dataset = self.get_input_dataset("bldgs2elec")
        bldgs2wter_dataset = self.get_input_dataset("bldgs2wter")

        if dislocation_data_type is not None and return_model is not None and testbed_name is not \
                None and bldgs2elec_dataset is not None and bldgs2wter_dataset is not None:
            dynamic_params = {
                "TYPE": dislocation_data_type,
                "RETURN": return_model,
                "TESTBED": testbed_name,
                "MAPPING": {'POWER': bldgs2elec_dataset.get_dataframe_from_csv(low_memory=False),
                            'WATER': bldgs2wter_dataset.get_dataframe_from_csv(low_memory=False)}
            }

        extra_commodity = self.get_parameter("extra_commodity")
        time_resource = self.get_parameter("time_resource")
        if time_resource is None:
            time_resource = True

        # save model or not; default to False
        save_model = self.get_parameter("save_model")
        if save_model is None:
            save_model = False

        action_result, cost_result, runtime_result = self.run_method(fail_sce_param, RC, layers, method=method,
                                                                     t_steps=t_steps,
                                                                     misc={'DYNAMIC_PARAMS': dynamic_params,
                                                                           'EXTRA_COMMODITY': extra_commodity,
                                                                           'TIME_RESOURCE': time_resource},
                                                                     save_model=save_model)

        self.set_result_csv_data("action", action_result, name="actions.csv")
        self.set_result_csv_data("cost", cost_result, name="costs.csv")
        self.set_result_csv_data("runtime", runtime_result, name="run_time.csv")

    def run_method(self, fail_sce_param, v_r, layers, method, t_steps=10, misc=None, save_model=False):
        """
        This function runs restoration analysis based on INDP or td-INDP for different numbers of resources.

        Args:
            fail_sce_param (dict): information about damage scenarios.
            v_r (list): number of resources, if this is a list of floats, each float is interpreted as a different
            total number of resources, and INDP
            is run given the total number of resources. If this is a list of lists of floats, each list is interpreted
            as fixed upper bounds on the number of resources each layer can use (same for all time steps).
            layers (list): List of layers.
            method (str): Algorithm type.
            t_steps (int): Number of time steps of the analysis.
            misc (dict): A dictionary that contains miscellaneous data needed for the analysis
            save_model (bool): Flag indicates if the model should be saved or not
        Returns:

        """

        # input files
        wf_repair_cost = self.get_input_dataset("wf_repair_cost").get_dataframe_from_csv(low_memory=False)
        wf_repair_cost['budget'] = wf_repair_cost['budget'].str.split(',')
        wf_repair_cost['repaircost'] = wf_repair_cost['repaircost'].str.split(',')
        epf_repair_cost = self.get_input_dataset("epf_repair_cost").get_dataframe_from_csv(low_memory=False)
        epf_repair_cost['budget'] = epf_repair_cost['budget'].str.split(',')
        epf_repair_cost['repaircost'] = epf_repair_cost['repaircost'].str.split(',')

        pipeline_restoration_time = self.get_input_dataset("pipeline_restoration_time").get_dataframe_from_csv(
            low_memory=False)
        pipeline_repair_cost = self.get_input_dataset("pipeline_repair_cost").get_dataframe_from_csv(low_memory=False)

        powerline_supply_demand_info = self.get_input_dataset("powerline_supply_demand_info").get_dataframe_from_csv(
            low_memory=False)
        epf_supply_demand_info = self.get_input_dataset("epf_supply_demand_info").get_dataframe_from_csv(
            low_memory=False)
        power_network = NetworkDataset.from_dataset(self.get_input_dataset("power_network"))
        power_arcs = power_network.links.get_dataframe_from_shapefile().merge(powerline_supply_demand_info, on="guid")
        power_nodes = power_network.nodes.get_dataframe_from_shapefile().merge(epf_supply_demand_info, on="guid")

        pipeline_supply_demand_info = self.get_input_dataset("pipeline_supply_demand_info").get_dataframe_from_csv(
            low_memory=False)
        wf_supply_demand_info = self.get_input_dataset("wf_supply_demand_info").get_dataframe_from_csv(low_memory=False)
        water_network = NetworkDataset.from_dataset(self.get_input_dataset("water_network"))
        water_arcs = water_network.links.get_dataframe_from_shapefile().merge(pipeline_supply_demand_info, on="guid")
        water_nodes = water_network.nodes.get_dataframe_from_shapefile()
        water_nodes = water_nodes.merge(wf_supply_demand_info, on="guid")

        interdep = self.get_input_dataset("interdep").get_dataframe_from_csv(low_memory=False)

        # get rid of distribution nodes
        wf_failure_state_df = self.get_input_dataset("wf_failure_state").get_dataframe_from_csv(
            low_memory=False).dropna()
        wf_damage_state_df = self.get_input_dataset("wf_damage_state").get_dataframe_from_csv(low_memory=False).dropna()
        pipeline_failure_state_df = self.get_input_dataset("pipeline_failure_state").get_dataframe_from_csv(
            low_memory=False).dropna()
        epf_failure_state_df = self.get_input_dataset("epf_failure_state").get_dataframe_from_csv(
            low_memory=False).dropna()
        epf_damage_state_df = self.get_input_dataset("epf_damage_state").get_dataframe_from_csv(
            low_memory=False).dropna()

        sample_range = self.get_parameter("sample_range")
        initial_node = INDPUtil.generate_intial_node_failure_state(wf_failure_state_df, epf_failure_state_df,
                                                                   water_nodes, power_nodes, sample_range)
        initial_link = INDPUtil.generate_intial_link_failure_state(pipeline_failure_state_df,
                                                                   water_arcs, power_arcs, sample_range)

        pop_dislocation = self.get_input_dataset("pop_dislocation").get_dataframe_from_csv(low_memory=False)

        wf_restoration_time = self.get_input_dataset("wf_restoration_time").get_dataframe_from_csv(low_memory=False)
        wf_restoration_time = wf_restoration_time.merge(wf_damage_state_df, on="guid")

        epf_restoration_time = self.get_input_dataset("epf_restoration_time").get_dataframe_from_csv(low_memory=False)
        epf_restoration_time = epf_restoration_time.merge(epf_damage_state_df, on="guid")

        dt_params_dataset = self.get_input_dataset("dt_params")
        if dt_params_dataset is not None:
            dt_params = dt_params_dataset.get_json_reader()
        else:
            dt_params = {'DS0': 1.00, 'DS1': 2.33, 'DS2': 2.49, 'DS3': 3.62, 'white': 0.78, 'black': 0.88,
                         'hispanic': 0.83, 'income': -0.00, 'insurance': 1.06}

        # results
        action_result = []
        cost_result = []
        runtime_result = []
        for v_i, v in enumerate(v_r):
            if method == 'INDP':
                params = {"NUM_ITERATIONS": t_steps, "OUTPUT_DIR": 'indp_results', "V": v,
                          "T": 1, 'L': layers, "ALGORITHM": "INDP"}
            elif method == 'TDINDP':
                params = {"NUM_ITERATIONS": t_steps, "OUTPUT_DIR": 'tdindp_results', "V": v,
                          "T": t_steps, 'L': layers, "ALGORITHM": "INDP"}
                if 'WINDOW_LENGTH' in misc.keys():
                    params["WINDOW_LENGTH"] = misc['WINDOW_LENGTH']
            else:
                raise ValueError('Wrong method name: ' + method + '. We currently only support INDP and TDINDP as  '
                                                                  'method name')

            params['EXTRA_COMMODITY'] = misc['EXTRA_COMMODITY']
            params['TIME_RESOURCE'] = misc['TIME_RESOURCE']
            params['DYNAMIC_PARAMS'] = misc['DYNAMIC_PARAMS']
            if misc['DYNAMIC_PARAMS']:
                params['OUTPUT_DIR'] = 'dp_' + params['OUTPUT_DIR']

            print('----Running for resources: ' + str(params['V']))
            for m in fail_sce_param['MAGS']:
                for i in fail_sce_param['SAMPLE_RANGE']:
                    params["SIM_NUMBER"] = i
                    params["MAGNITUDE"] = m

                    print('---Running Magnitude ' + str(m) + ' sample ' + str(i) + '...')
                    if params['TIME_RESOURCE']:
                        print('Computing repair times...')

                        wf_repair_cost_sample = wf_repair_cost.copy()
                        wf_repair_cost_sample["budget"] = wf_repair_cost_sample['budget'].apply(lambda x: float(x[i]))
                        wf_repair_cost_sample["repaircost"] = \
                            wf_repair_cost_sample['repaircost'].apply(lambda x: float(x[i]))

                        epf_repair_cost_sample = epf_repair_cost.copy()
                        epf_repair_cost_sample["budget"] = epf_repair_cost_sample['budget'].apply(lambda x: float(x[i]))
                        epf_repair_cost_sample["repaircost"] = \
                            epf_repair_cost_sample['repaircost'].apply(lambda x: float(x[i]))

                        # logic to read repair time
                        wf_restoration_time_sample = pd.DataFrame()
                        for index, row in wf_restoration_time.iterrows():
                            failure_state = int(row["sample_damage_states"].split(",")[i].split("_")[1])  # DS_0,1,2,3,4
                            if failure_state == 0:
                                repairtime = 0
                            else:
                                repairtime = row["PF_" + str(failure_state - 1)]
                            wf_restoration_time_sample = pd.concat([wf_restoration_time_sample,
                                                                    pd.DataFrame([{"guid": row["guid"],
                                                                                  "repairtime": repairtime}])],
                                                                   ignore_index=True)
                        epf_restoration_time_sample = pd.DataFrame()
                        for index, row in epf_restoration_time.iterrows():
                            failure_state = int(row["sample_damage_states"].split(",")[i].split("_")[1])  # DS_0,1,2,3,4
                            if failure_state == 0:
                                repairtime = 0
                            else:
                                repairtime = row["PF_" + str(failure_state - 1)]
                            epf_restoration_time_sample = pd.concat([epf_restoration_time_sample,
                                                                     pd.DataFrame([{"guid": row["guid"],
                                                                                   "repairtime":repairtime}])],
                                                                    ignore_index=True)

                        water_nodes, water_arcs, power_nodes, power_arcs = \
                            INDPUtil.time_resource_usage_curves(power_arcs, power_nodes, water_arcs, water_nodes,
                                                                wf_restoration_time_sample, wf_repair_cost_sample,
                                                                pipeline_restoration_time, pipeline_repair_cost,
                                                                epf_restoration_time_sample, epf_repair_cost_sample)

                    print("Initializing network...")
                    params["N"] = INDPUtil.initialize_network(power_nodes, power_arcs, water_nodes, water_arcs,
                                                              interdep, extra_commodity=params["EXTRA_COMMODITY"])

                    if params['DYNAMIC_PARAMS']:
                        print("Computing dynamic demand based on dislocation data...")
                        dyn_dmnd = DislocationUtil.create_dynamic_param(params, pop_dislocation, dt_params,
                                                                        N=params["N"], T=params["NUM_ITERATIONS"])
                        params['DYNAMIC_PARAMS']['DEMAND_DATA'] = dyn_dmnd

                    if fail_sce_param['TYPE'] == 'from_csv':
                        InfrastructureUtil.add_from_csv_failure_scenario(params["N"], sample=i,
                                                                         initial_node=initial_node,
                                                                         initial_link=initial_link)
                    else:
                        raise ValueError('Wrong failure scenario data type.')

                    if params["ALGORITHM"] == "INDP":
                        indp_results = self.run_indp(params, layers=params['L'], controlled_layers=params['L'],
                                                     T=params["T"], save_model=save_model, print_cmd_line=False,
                                                     co_location=False)
                        for t in indp_results.results:
                            actions = indp_results[t]['actions']
                            costs = indp_results[t]['costs']
                            runtimes = indp_results[t]['run_time']
                            for a in actions:
                                action_result.append({
                                    "RC": str(v_i),
                                    "layers": 'L' + str(len(layers)),
                                    "magnitude": 'm' + str(m),
                                    "sample_num": str(i),
                                    "t": str(t),
                                    "action": a,
                                })

                            runtime_result.append({
                                "RC": str(v_i),
                                "layers": 'L' + str(len(layers)),
                                "magnitude": 'm' + str(m),
                                "sample_num": str(i),
                                "t": str(t),
                                "runtime": runtimes,
                            })

                            cost_result.append({
                                "RC": str(v_i),
                                "layers": 'L' + str(len(layers)),
                                "magnitude": 'm' + str(m),
                                "sample_num": str(i),
                                "t": str(t),
                                "Space Prep": str(costs["Space Prep"]),
                                "Arc": str(costs["Arc"]),
                                "Node": str(costs["Node"]),
                                "Over Supply": str(costs["Over Supply"]),
                                "Under Supply": str(costs["Under Supply"]),
                                "Flow": str(costs["Flow"]),
                                "Total": str(costs["Total"]),
                                "Under Supply Perc": str(costs["Under Supply Perc"]),
                            })
                    else:
                        raise ValueError('Wrong algorithm type.')

        return action_result, cost_result, runtime_result

    def run_indp(self, params, layers=None, controlled_layers=None, functionality=None, T=1, save=True, suffix="",
                 forced_actions=False, save_model=False, print_cmd_line=True, co_location=True):
        """
        This function runs iINDP (T=1) or td-INDP for a given number of time steps and input parameters.

        Args:
            params (dict): Parameters that are needed to run the INDP optimization.
            layers (list): List of layers in the interdependent network. The default is 'None', which sets the list
            to [1, 2, 3].
            controlled_layers (list): List of layers that are included in the analysis. The default is 'None',
            which  sets the list equal to layers.
            functionality (dict): This dictionary is used to assign functionality values elements in the network
            before  the analysis starts. The
            default is 'None'.
            T (int): Number of time steps to optimize over. T=1 shows an iINDP analysis, and T>1 shows a td-INDP.
            The default is 1.
            TODO save & suffice aare not exposed to outside should remove it
            save (bool): If the results should be saved to file. The default is True.
            suffix (str): The suffix that should be added to the output files when saved. The default is ''.

            forced_actions (bool): If True, the optimizer is forced to repair at least one element in each time step.
            The default is False.
            TODO expose this parameter
            save_model (bool): If the optimization model should be saved to file. The default is False.
            TODO expose this parameter
            print_cmd_line (bool): If full information about the analysis should be written to the console. The default
            is True.
            TODO expose this parameter
            co_location (bool): If co-location and geographical interdependency should be considered in the analysis.
            The default is True.

        Returns:
             indp_results (INDPResults): `~indputils.INDPResults` object containing the optimal restoration decisions.

        """
        # Initialize failure scenario.
        global original_N
        if functionality is None:
            functionality = {}
        if layers is None:
            layers = [1, 2, 3]
        if controlled_layers is None:
            controlled_layers = layers

        interdependent_net = params["N"]
        if "NUM_ITERATIONS" not in params:
            params["NUM_ITERATIONS"] = 1

        out_dir_suffix_res = INDPUtil.get_resource_suffix(params)
        indp_results = INDPResults(params["L"])
        if T == 1:
            print("--Running INDP (T=1) or iterative INDP.")
            if print_cmd_line:
                print("Num iters=", params["NUM_ITERATIONS"])

            # Run INDP for 1 time step (original INDP).
            output_dir = params["OUTPUT_DIR"] + '_L' + str(len(layers)) + '_m' + str(
                params["MAGNITUDE"]) + "_v" + out_dir_suffix_res
            # Initial calculations.
            if params['DYNAMIC_PARAMS']:
                original_N = copy.deepcopy(interdependent_net)  # !!! deepcopy
                DislocationUtil.dynamic_parameters(interdependent_net, original_N, 0,
                                                   params['DYNAMIC_PARAMS']['DEMAND_DATA'])
            v_0 = {x: 0 for x in params["V"].keys()}
            results = self.indp(interdependent_net, v_0, 1, layers, controlled_layers=controlled_layers,
                                functionality=functionality, co_location=co_location)
            indp_results = results[1]
            if save_model:
                INDPUtil.save_indp_model_to_file(results[0], output_dir + "/Model", 0)
            for i in range(params["NUM_ITERATIONS"]):
                print("-Time Step (iINDP)", i + 1, "/", params["NUM_ITERATIONS"])
                if params['DYNAMIC_PARAMS']:
                    DislocationUtil.dynamic_parameters(interdependent_net, original_N, i + 1,
                                                       params['DYNAMIC_PARAMS']['DEMAND_DATA'])
                results = self.indp(interdependent_net, params["V"], T, layers, controlled_layers=controlled_layers,
                                    co_location=co_location, functionality=functionality)
                indp_results.extend(results[1], t_offset=i + 1)
                if save_model:
                    INDPUtil.save_indp_model_to_file(results[0], output_dir + "/Model", i + 1)
                # Modify network to account for recovery and calculate components.
                INDPUtil.apply_recovery(interdependent_net, indp_results, i + 1)
        else:
            # td-INDP formulations. Includes "DELTA_T" parameter for sliding windows to increase efficiency.
            # Edit 2/8/16: "Sliding window" now overlaps.
            num_time_windows = 1
            time_window_length = T
            if "WINDOW_LENGTH" in params:
                time_window_length = params["WINDOW_LENGTH"]
                num_time_windows = T
            output_dir = params["OUTPUT_DIR"] + '_L' + str(len(layers)) + "_m" + str(
                params["MAGNITUDE"]) + "_v" + out_dir_suffix_res

            print("Running td-INDP (T=" + str(T) + ", Window size=" + str(time_window_length) + ")")
            # Initial percolation calculations.
            v_0 = {x: 0 for x in params["V"].keys()}
            results = self.indp(interdependent_net, v_0, 1, layers, controlled_layers=controlled_layers,
                                functionality=functionality, co_location=co_location)
            indp_results = results[1]
            if save_model:
                INDPUtil.save_indp_model_to_file(results[0], output_dir + "/Model", 0)
            for n in range(num_time_windows):
                print("-Time window (td-INDP)", n + 1, "/", num_time_windows)
                functionality_t = {}
                # Slide functionality matrix according to sliding time window.
                if functionality:
                    for t in functionality:
                        if t in range(n, time_window_length + n + 1):
                            functionality_t[t - n] = functionality[t]
                    if len(functionality_t) < time_window_length + 1:
                        diff = time_window_length + 1 - len(functionality_t)
                        max_t = max(functionality_t.keys())
                        for d in range(diff):
                            functionality_t[max_t + d + 1] = functionality_t[max_t]
                # Run td-INDP.
                results = self.indp(interdependent_net, params["V"], time_window_length + 1, layers,
                                    controlled_layers=controlled_layers, functionality=functionality_t,
                                    co_location=co_location)
                if save_model:
                    INDPUtil.save_indp_model_to_file(results[0], output_dir + "/Model", n + 1)
                if "WINDOW_LENGTH" in params:
                    indp_results.extend(results[1], t_offset=n + 1, t_start=1, t_end=2)
                    # Modify network for recovery actions and calculate components.
                    INDPUtil.apply_recovery(interdependent_net, results[1], 1)
                else:
                    indp_results.extend(results[1], t_offset=0)
                    for t in range(1, T):
                        # Modify network to account for recovery actions.
                        INDPUtil.apply_recovery(interdependent_net, indp_results, t)
        # Save results of current simulation.
        if save:
            if not os.path.exists(output_dir + '/agents'):
                os.makedirs(output_dir + '/agents')
            indp_results.to_csv_layer(output_dir + '/agents', params["SIM_NUMBER"], suffix=suffix)

        return indp_results

    def indp(self, N, v_r, T=1, layers=None, controlled_layers=None, functionality=None, fixed_nodes=None,
             print_cmd=True, co_location=True):
        """
        INDP optimization problem in Pyomo. It also solves td-INDP if T > 1.

        Parameters
        ----------
        N : :class:`~infrastructure.InfrastructureNetwork`
            An InfrastructureNetwork instance.
        v_r : dict
            Dictionary of the number of resources of different types in the analysis.
            If the value is a scalar for a type, it shows the total number of resources of that type for all layers.
            If the value is a list for a type, it shows the total number of resources of that type given to each layer.
        T : int, optional
            Number of time steps to optimize over. T=1 shows an iINDP analysis, and T>1 shows a td-INDP. The default is
            1.
        layers : list, optional
            Layer IDs in N included in the optimization.
        controlled_layers : list, optional
            Layer IDs that can be recovered in this optimization. Used for decentralized optimization. The default is
            None.
        functionality : dict, optional
            Dictionary of nodes to functionality values for non-controlled nodes.
            Used for decentralized optimization. The default is None.
        fixed_nodes : dict, optional
            It fixes the functionality of given elements to a given value. The default is None.
        print_cmd : bool, optional
            If true, analysis information is written to the console. The default is True.
        co_location : bool, optional
            If false, exclude geographical interdependency from the optimization. The default is True.
        Returns
        -------
        : list
        A list of the form ``[m, results]`` for a successful optimization where m is the optimization model and
            results is a :class:`~indputils.INDPResults` object generated using  :func:`collect_results`.
            If :envvar:`solution_pool` is set to a number, the function returns ``[m, results,  sol_pool_results]``
            where `sol_pool_results` is dictionary of solution that should be retrieved from the optimizer in
            addition to the optimal one collected using :func:`collect_solution_pool`.

        """
        if functionality is None:
            functionality = {}
        if layers is None:
            layers = [1, 2, 3]
        if controlled_layers is None:
            controlled_layers = layers

        start_time = time.time()

        m = pyo.ConcreteModel()
        m.T = T
        m.N = N
        m.v_r = v_r
        m.functionality = functionality

        '''Sets and Dictionaries'''
        g_prime_nodes = [n[0] for n in N.G.nodes(data=True) if n[1]['data']['inf_data'].net_id in layers]
        g_prime = N.G.subgraph(g_prime_nodes)
        # Nodes in controlled network.
        m.n_hat_nodes = pyo.Set(
            initialize=[n[0] for n in g_prime.nodes(data=True) if n[1]['data']['inf_data'].net_id in controlled_layers])
        m.n_hat = g_prime.subgraph(m.n_hat_nodes.ordered_data())
        # Damaged nodes in controlled network.
        m.n_hat_prime_nodes = pyo.Set(
            initialize=[n[0] for n in m.n_hat.nodes(data=True) if n[1]['data']['inf_data'].repaired == 0.0])
        n_hat_prime = [n for n in m.n_hat.nodes(data=True) if n[1]['data']['inf_data'].repaired == 0.0]
        # Arcs in controlled network.
        m.a_hat = pyo.Set(
            initialize=[(u, v) for u, v, a in g_prime.edges(data=True) if
                        a['data']['inf_data'].layer in controlled_layers])
        # Damaged arcs in whole network
        m.a_prime = pyo.Set(
            initialize=[(u, v) for u, v, a in g_prime.edges(data=True) if a['data']['inf_data'].functionality == 0.0])
        a_prime = [(u, v, a) for u, v, a in g_prime.edges(data=True) if a['data']['inf_data'].functionality == 0.0]
        # Damaged arcs in controlled network.
        m.a_hat_prime = pyo.Set(initialize=[(u, v) for u, v, _ in a_prime if m.n_hat.has_node(u)
                                            and m.n_hat.has_node(v)])
        a_hat_prime = [(u, v, a) for u, v, a in a_prime if m.n_hat.has_node(u) and m.n_hat.has_node(v)]
        # Sub-spaces
        m.S = pyo.Set(initialize=N.S)

        # Populate interdependencies. Add nodes to N' if they currently rely on a non-functional node.
        m.interdep_nodes = {}
        for u, v, a in g_prime.edges(data=True):
            if not functionality:
                if a['data']['inf_data'].is_interdep and g_prime.nodes[u]['data']['inf_data'].functionality == 0.0:
                    # print "Dependency edge goes from:",u,"to",v
                    if v not in m.interdep_nodes:
                        m.interdep_nodes[v] = []
                    m.interdep_nodes[v].append((u, a['data']['inf_data'].gamma))
            else:
                # Should populate m.n_hat with layers that are controlled. Then go through m.n_hat.edges(data=True)
                # to find interdependencies.
                for t in range(T):
                    if t not in m.interdep_nodes:
                        m.interdep_nodes[t] = {}
                    if m.n_hat.has_node(v) and a['data']['inf_data'].is_interdep:
                        if functionality[t][u] == 0.0:
                            if v not in m.interdep_nodes[t]:
                                m.interdep_nodes[t][v] = []
                            m.interdep_nodes[t][v].append((u, a['data']['inf_data'].gamma))

        '''Variables'''
        m.time_step = pyo.Set(initialize=range(T))
        # Add geographical space variables.
        if co_location:
            m.S_ids = pyo.Set(initialize=[s.id for s in m.S.value])
            m.z = pyo.Var(m.S_ids, m.time_step, domain=pyo.Binary)
        # Add functionality binary variables for each node in N'.
        m.w = pyo.Var(m.n_hat_nodes, m.time_step, domain=pyo.Binary)
        if T > 1:
            m.w_tilde = pyo.Var(m.n_hat_nodes, m.time_step, domain=pyo.Binary)
        # Add functionality binary variables for each arc in A'.
        m.y = pyo.Var(m.a_hat_prime, m.time_step, domain=pyo.Binary)
        if T > 1:
            m.y_tilde = pyo.Var(m.a_hat_prime, m.time_step, domain=pyo.Binary)
        # Add variables considering extra commodity in addition to the base one
        node_com_idx = []
        for n, d in m.n_hat.nodes(data=True):
            node_com_idx.append((n, 'b'))
            for key, val in d['data']['inf_data'].extra_com.items():
                node_com_idx.append((n, key))
        arc_com_idx = []
        for u, v, a in m.n_hat.edges(data=True):
            arc_com_idx.append((u, v, 'b'))
            for key, val in a['data']['inf_data'].extra_com.items():
                arc_com_idx.append((u, v, key))
        # Add over/under-supply variables for each node.
        m.delta_p = pyo.Var(node_com_idx, m.time_step, domain=pyo.NonNegativeReals)
        m.delta_m = pyo.Var(node_com_idx, m.time_step, domain=pyo.NonNegativeReals)
        # Add flow variables for each arc. (main commodity)
        m.x = pyo.Var(arc_com_idx, m.time_step, domain=pyo.NonNegativeReals)

        # Fix node values
        if fixed_nodes:
            for key, val in fixed_nodes.items():
                if m.T == 1:
                    m.w[key].fix(val)
                else:
                    m.w_tilde[key].fix(val)

        '''Populate objective function'''
        obj_func = 0
        for t in range(m.T):
            if co_location:
                for s in N.S:
                    obj_func += s.cost * m.z[s.id, t]
            for u, v, a in a_hat_prime:
                if T == 1:
                    obj_func += (float(a['data']['inf_data'].reconstruction_cost) / 2.0) * m.y[u, v, t]
                else:
                    obj_func += (float(a['data']['inf_data'].reconstruction_cost) / 2.0) * m.y_tilde[u, v, t]
            for n, d in n_hat_prime:
                if T == 1:
                    obj_func += d['data']['inf_data'].reconstruction_cost * m.w[n, t]
                else:
                    obj_func += d['data']['inf_data'].reconstruction_cost * m.w_tilde[n, t]
            for n, d in m.n_hat.nodes(data=True):
                obj_func += d['data']['inf_data'].oversupply_penalty * m.delta_p[n, 'b', t]
                obj_func += d['data']['inf_data'].undersupply_penalty * m.delta_m[n, 'b', t]
                for layer, val in d['data']['inf_data'].extra_com.items():
                    obj_func += val['oversupply_penalty'] * m.delta_p[n, layer, t]
                    obj_func += val['undersupply_penalty'] * m.delta_m[n, layer, t]
            for u, v, a in m.n_hat.edges(data=True):
                obj_func += a['data']['inf_data'].flow_cost * m.x[u, v, 'b', t]
                for layer, val in a['data']['inf_data'].extra_com.items():
                    obj_func += val['flow_cost'] * m.x[u, v, layer, t]
        m.Obj = pyo.Objective(rule=obj_func, sense=pyo.minimize)

        '''Constraints'''
        # Time-dependent constraints.
        if m.T > 1:
            m.initial_state_node = pyo.Constraint(m.n_hat_prime_nodes,
                                                  rule=INDPUtil.initial_state_node_rule,
                                                  doc='Initialstate at node')
            m.initial_state_arc = pyo.Constraint(m.a_hat_prime, rule=INDPUtil.initial_state_arc_rule,
                                                 doc='Initial state at arc')
        m.time_dependent_node = pyo.Constraint(m.n_hat_prime_nodes,
                                               m.time_step,
                                               rule=INDPUtil.time_dependent_node_rule,
                                               doc='Time dependent recovery constraint at node')
        m.time_dependent_arc = pyo.Constraint(m.a_hat_prime,
                                              m.time_step,
                                              rule=INDPUtil.time_dependent_arc_rule,
                                              doc='Time dependent recovery constraint at arc')
        # Enforce a_i,j to be fixed if a_j,i is fixed (and vice versa).
        m.arc_equality = pyo.Constraint(m.a_hat_prime,
                                        m.time_step,
                                        rule=INDPUtil.arc_equality_rule,
                                        doc='Arc reconstruction equality')

        # Conservation of flow constraint. (2) in INDP paper.
        m.flow_conserv_node = pyo.Constraint(m.delta_p_index_0,
                                             m.time_step,
                                             rule=INDPUtil.flow_conserv_node_rule,
                                             doc='Flow conservation')
        # Flow functionality constraints.
        m.flow_in_functionality = pyo.Constraint(m.a_hat, m.time_step, rule=INDPUtil.flow_in_functionality_rule,
                                                 doc='Flow In Functionality')
        m.flow_out_functionality = pyo.Constraint(m.a_hat, m.time_step, rule=INDPUtil.flow_out_functionality_rule,
                                                  doc='Flow Out Functionality')
        m.flow_arc_functionality = pyo.Constraint(m.a_hat, m.time_step, rule=INDPUtil.flow_arc_functionality_rule,
                                                  doc='Flow Arc Functionality')
        # Resource availability constraints.
        m.resource = pyo.Constraint(list(m.v_r.keys()), m.time_step, rule=INDPUtil.resource_rule,
                                    doc='Resource availability')
        # Interdependency constraints
        m.interdependency = pyo.Constraint(m.n_hat_nodes, m.time_step, rule=INDPUtil.interdependency_rule,
                                           doc='Interdependency')
        # Geographic space constraints
        if co_location:
            m.node_geographic_space = pyo.Constraint(m.S_ids, m.n_hat_prime_nodes, m.time_step,
                                                     rule=INDPUtil.node_geographic_space_rule,
                                                     doc='Node Geographic space')
            m.arc_geographic_space = pyo.Constraint(m.S_ids, m.a_hat_prime, m.time_step,
                                                    rule=INDPUtil.arc_geographic_space_rule,
                                                    doc='Arc Geographic space')

        '''Solve'''
        num_cont_vars = len([v for v in m.component_data_objects(pyo.Var) if v.domain == pyo.NonNegativeReals])
        num_integer_vars = len([v for v in m.component_data_objects(pyo.Var) if v.domain == pyo.Binary])

        solver_engine = self.get_parameter("solver_engine")
        if solver_engine is None:
            solver_engine = "scip"

        solver_path = self.get_parameter("solver_path")
        if solver_path is None:
            solver_path = pyglobals.SCIP_PATH

        solver_time_limit = self.get_parameter("solver_time_limit")

        print(
            "Solving... using %s solver (%d cont. vars, %d binary vars)" %
            (solver_engine, num_cont_vars, num_integer_vars))

        if solver_engine == "gurobi":
            solver = SolverFactory(solver_engine, timelimit=solver_time_limit)
        else:
            solver = SolverFactory(solver_engine, timelimit=solver_time_limit, executable=solver_path)

        solution = solver.solve(m)
        run_time = time.time() - start_time

        # Save results.
        if solution.solver.termination_condition in [TerminationCondition.optimal, TerminationCondition.maxTimeLimit]:
            if solution.solver.termination_condition == TerminationCondition.maxTimeLimit:
                print('\nOptimizer time limit, gap = %1.3f\n' % solution.a.solution(0).gap)
            results = INDPUtil.collect_results(m, controlled_layers, coloc=co_location)
            results.add_run_time(t, run_time)
            return [m, results]
        else:
            log_infeasible_constraints(m, log_expression=True, log_variables=True)
            if solution.solver.termination_condition == TerminationCondition.infeasible:
                print(solution.solver.termination_condition, ": SOLUTION NOT FOUND. (Check data and/or violated "
                                                             "constraints in the infeasible_model.log).")
            sys.exit()

    def get_spec(self):
        return {
            'name': 'INDP',
            'description': 'Interdependent Network Design Problem that models the restoration',
            'input_parameters': [
                {
                    'id': 'network_type',
                    'required': True,
                    'description': 'type of the network, which is set to `from_csv` for Seaside networks. '
                                   'e.g. from_csv, incore',
                    'type': str
                },
                {
                    'id': 'MAGS',
                    'required': True,
                    'description': 'sets the earthquake return period.',
                    'type': list
                },
                {
                    'id': 'sample_range',
                    'required': True,
                    'description': 'sets the range of sample scenarios to be analyzed',
                    'type': range
                },
                {
                    'id': 'dislocation_data_type',
                    'required': False,
                    'description': 'type of the dislocation data.',
                    'type': str
                },
                {
                    'id': 'return_model',
                    'required': False,
                    'description': 'type of the model for the return of the dislocated population. '
                                   'Options: *step_function* and *linear*.',
                    'type': str
                },
                {
                    'id': 'testbed_name',
                    'required': False,
                    'description': 'sets the name of the testbed in analysis',
                    'type': str
                },
                {
                    'id': 'extra_commodity',
                    'required': True,
                    'description': 'multi-commodity parameters dict',
                    'type': dict
                },
                {
                    'id': 'RC',
                    'required': True,
                    'description': 'list of resource caps or the number of available resources in each step of the '
                                   'analysis. Each item of the list is a dictionary whose items show the type of '
                                   'resource and the available number of that type of resource. For example: '
                                   '* If `network_type`=*from_csv*, you have two options:* if, for example, '
                                   '`R_c`= [{"budget": 3}, {"budget": 6}], then the analysis is done for the cases '
                                   'when there are 3 and 6 resources available of type "budget" '
                                   '(total resource assignment).* if, for example, `R_c`= [{"budget": {1:1, 2:1}}, '
                                   '{"budget": {1:1, 2:2}}, {"budget": {1:3, 2:3}}] and given there are 2 layers,'
                                   ' then the analysis is done for the case where each layer gets 1 resource of '
                                   'type "budget", AND the case where layer 1 gets 1 and layer 2 gets 2 resources of '
                                   'type "budget", AND the case where each layer gets 3 resources of type '
                                   '"budget" (Prescribed resource for each layer).',
                    'type': list
                },
                {
                    'id': 'layers',
                    'required': True,
                    'description': 'list of layers in the analysis',
                    'type': list
                },
                {
                    'id': 'method',
                    'required': True,
                    'description': 'There are two choices of method: 1. `INDP`: runs Interdependent Network '
                                   'Restoration Problem (INDP). 2. `TDINDP`: runs time-dependent INDP (td-INDP).  In '
                                   'both cases, if "TIME_RESOURCE" is True, then the repair time for each element '
                                   'is considered in devising the restoration plans',
                    'type': str,
                },
                {
                    'id': 't_steps',
                    'required': False,
                    'description': 'Number of time steps of the analysis',
                    'type': int
                },
                {
                    'id': 'time_resource',
                    'required': False,
                    'description': 'if TIME_RESOURCE is True, then the repair time for each element is '
                                   'considered in devising the restoration plans',
                    'type': bool
                },
                {
                    'id': 'save_model',
                    'required': False,
                    'description': 'If the optimization model should be saved to file. The default is False.',
                    'type': bool
                },
                {
                    'id': 'solver_engine',
                    'required': False,
                    'description': "Solver to use for optimization model. Such as gurobi/glpk/scip, default to scip.",
                    'type': str
                },
                {
                    'id': 'solver_path',
                    'required': False,
                    'description': "Solver to use for optimization model. Such as gurobi/glpk/scip, default to scip.",
                    'type': str
                },
                {
                    'id': 'solver_time_limit',
                    'required': False,
                    'description': "solver time limit in seconds",
                    'type': int
                },
            ],
            'input_datasets': [
                {
                    "id": "wf_repair_cost",
                    "required": True,
                    "description": "repair cost for each water facility",
                    "type": ["incore:repairCost"]
                },
                {
                    'id': 'wf_restoration_time',
                    "required": True,
                    'description': 'recording repair time at certain functionality recovery for each class '
                                   'and limit state.',
                    'type': ['incore:waterFacilityRepairTime']
                },
                {
                    "id": "epf_repair_cost",
                    "required": True,
                    "description": "repair cost for each electric power facility",
                    "type": ["incore:repairCost"]
                },
                {
                    'id': 'epf_restoration_time',
                    "required": True,
                    'description': 'recording repair time at certain functionality recovery for each class '
                                   'and limit state.',
                    'type': ['incore:epfRepairTime']
                },
                {
                    "id": "pipeline_repair_cost",
                    "required": True,
                    "description": "repair cost for each pipeline",
                    "type": ["incore:pipelineRepairCost"]
                },
                {
                    'id': 'pipeline_restoration_time',
                    "required": True,
                    'description': 'pipeline restoration times',
                    'type': ['incore:pipelineRestorationVer1']
                },
                {
                    'id': 'power_network',
                    'required': True,
                    'description': 'EPN Network Dataset',
                    'type': ['incore:epnNetwork'],
                },
                {
                    'id': 'water_network',
                    'required': True,
                    'description': 'Water Network Dataset',
                    'type': ['incore:waterNetwork'],
                },
                {
                    'id': 'powerline_supply_demand_info',
                    'required': True,
                    'description': 'Supply and demand information for powerlines',
                    'type': ['incore:powerLineSupplyDemandInfo'],
                },
                {
                    'id': 'epf_supply_demand_info',
                    'required': True,
                    'description': 'Supply and demand information for epfs',
                    'type': ['incore:epfSupplyDemandInfo'],
                },
                {
                    'id': 'wf_supply_demand_info',
                    'required': True,
                    'description': 'Supply and demand information for water facilities',
                    'type': ['incore:waterFacilitySupplyDemandInfo'],
                },
                {
                    'id': 'pipeline_supply_demand_info',
                    'required': True,
                    'description': 'Supply and demand information for water pipelines',
                    'type': ['incore:pipelineSupplyDemandInfo'],
                },
                {
                    "id": "interdep",
                    "required": True,
                    "description": "Interdepenency between water and electric power facilities",
                    "type": ["incore:interdep"]
                },
                {
                    "id": "wf_failure_state",
                    "required": True,
                    "description": "MCS failure state of water facilities",
                    "type": ["incore:sampleFailureState"]
                },
                {
                    "id": "wf_damage_state",
                    "required": True,
                    "description": "MCS damage state of water facilities",
                    "type": ["incore:sampleDamageState"]
                },
                {
                    "id": "pipeline_failure_state",
                    "required": True,
                    "description": "failure state of pipeline from pipeline functionality",
                    "type": ["incore:sampleFailureState"]
                },
                {
                    "id": "epf_failure_state",
                    "required": True,
                    "description": "MCS failure state of electric power facilities",
                    "type": ["incore:sampleFailureState"]
                },
                {
                    "id": "epf_damage_state",
                    "required": True,
                    "description": "MCS damage state of electric power facilities",
                    "type": ["incore:sampleDamageState"]
                },
                {
                    "id": "dt_params",
                    "required": False,
                    "description": "Parameters for population dislocation time",
                    "type": ["incore:dTParams"]
                },
                {
                    "id": "pop_dislocation",
                    "required": True,
                    "description": "Population dislocation output",
                    "type": ["incore:popDislocation"]
                },
                {
                    "id": "bldgs2elec",
                    "required": False,
                    "description": "relation between building and electric power facility",
                    "type": ["incore:bldgs2elec"]
                },
                {
                    "id": "bldgs2wter",
                    "required": False,
                    "description": "relation between building and water facility",
                    "type": ["incore:bldgs2wter"]
                },
            ],
            'output_datasets': [
                {
                    'id': 'action',
                    'parent_type': '',
                    'description': 'Restoration action plans',
                    'type': 'incore:indpAction'
                },
                {
                    'id': 'cost',
                    'parent_type': '',
                    'description': 'Restoration cost plans',
                    'type': 'incore:indpCost'
                },
                {
                    'id': 'runtime',
                    'parent_type': '',
                    'description': 'Restoration runtime plans',
                    'type': 'incore:indpRuntime'
                }
            ]
        }



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import json

from pyincore.models.dfr3curve import DFR3Curve
from pyincore.utils.analysisutil import AnalysisUtil


class FragilityCurveSet:
    """A class for fragility curves.

    Args:
        metadata (dict): fragility curve metadata.

    Raises:
        ValueError: Raised if there are unsupported number of fragility curves
        or if missing a key curve field.

    """

    def __init__(self, metadata):
        self.id = metadata["id"] if "id" in metadata else ""
        self.description = metadata['description'] if "description" in metadata else ""
        self.authors = ", ".join(metadata['authors']) if "authors" in metadata else ""
        self.paper_reference = str(metadata["paperReference"]) if "paperReference" in metadata else ""
        self.creator = metadata["creator"] if "creator" in metadata else ""
        self.demand_types = metadata["demandTypes"]
        self.demand_units = metadata["demandUnits"]
        self.result_type = metadata["resultType"]
        self.result_unit = metadata["resultUnit"] if "resultUnit" in metadata else ""
        self.hazard_type = metadata['hazardType']
        self.inventory_type = metadata['inventoryType']
        self.curve_parameters = {}
        self.fragility_curves = []

        if 'curveParameters' in metadata.keys():
            self.curve_parameters = metadata["curveParameters"]

        if 'fragilityCurves' in metadata.keys():
            for fragility_curve in metadata["fragilityCurves"]:
                self.fragility_curves.append(DFR3Curve(fragility_curve))
        else:
            raise ValueError("Cannot create dfr3 curve object. Missing key field: fragilityCurves.")

    @classmethod
    def from_json_str(cls, json_str):
        """Get dfr3set from json string.

        Args:
            json_str (str): JSON of the Dataset.

        Returns:
            obj: dfr3set from JSON.

        """
        return cls(json.loads(json_str))

    @classmethod
    def from_json_file(cls, file_path):
        """Get dfr3set from the file.

        Args:
            file_path (str): json file path that holds the definition of a dfr3 curve.

        Returns:
            obj: dfr3set from file.

        """
        with open(file_path, "r") as f:
            instance = cls(json.load(f))

        return instance

    def calculate_limit_state(self, hazard_values: dict = {},
                              inventory_type: str = "building",
                              **kwargs):
        """WIP computation of limit state probabilities accounting for custom expressions.

        Args:
            hazard_values (dict): A dictionary with hazard values to compute probability.
            inventory_type (str): An inventory type.
            **kwargs: Keyword arguments.

        Returns:
            OrderedDict: Limit state probabilities.

        """
        output = FragilityCurveSet._initialize_limit_states(inventory_type)
        limit_state = list(output.keys())
        index = 0

        if len(self.fragility_curves) <= 4:
            for fragility_curve in self.fragility_curves:
                probability = fragility_curve.solve_curve_expression(hazard_values,
                                                                     self.curve_parameters,
                                                                     **kwargs)
                output[limit_state[index]] = AnalysisUtil.update_precision(probability)  # round to default digits
                index += 1
        else:
            raise ValueError("We can only handle fragility curves with less than 4 limit states.")

        return output

    def calculate_damage_interval(self, damage, hazard_type="earthquake", inventory_type: str = "building"):
        """

        Args:
            damage (list): A list of limit states.
            hazard_type (str): A string describing the hazard being evaluated.
            inventory_type (str): A string describing the type of element being evaluated.

        Returns:
            list: LS-to-DS mapping

        """
        # Organize conceptually per LS-to-DS mapping , then by event, then by structure and by count
        # This may help keep track of scientific requirements also.

        ls_ds_dspatcher = {
            # 1 LS to 4 DS
            ("hurricane", "building", 1): FragilityCurveSet._1ls_to_4ds,
            ("hurricane", "electric_facility", 1): FragilityCurveSet._1ls_to_4ds,
            # 1 LS to 5 DS
            ("hurricane", "road", 1): FragilityCurveSet._1ls_to_5ds,
            ("hurricane", "bridge", 1): FragilityCurveSet._1ls_to_5ds,
            # 3 LS to 4 DS
            ("earthquake", "building", 3): FragilityCurveSet._3ls_to_4ds,
            ("earthquake+tsunami", "building", 3): FragilityCurveSet._3ls_to_4ds,
            ("tornado", "building", 3): FragilityCurveSet._3ls_to_4ds,
            ("hurricane", "building", 3): FragilityCurveSet._3ls_to_4ds,
            ("flood", "building", 3): FragilityCurveSet._3ls_to_4ds,
            ("tsunami", "building", 3): FragilityCurveSet._3ls_to_4ds,
            # 4 LS to 5 DS
            ("earthquake", "bridge", 4): FragilityCurveSet._4ls_to_5ds,
            ("earthquake", "pipeline", 4): FragilityCurveSet._4ls_to_5ds,
            ("earthquake", "road", 4): FragilityCurveSet._4ls_to_5ds,
            ("earthquake", "water_facility", 4): FragilityCurveSet._4ls_to_5ds,
            ("earthquake", "electric_facility", 4): FragilityCurveSet._4ls_to_5ds,
            ("tornado", "bridge", 4): FragilityCurveSet._4ls_to_5ds,
            ("tornado", "electric_facility", 4): FragilityCurveSet._4ls_to_5ds,
            ("flood", "bridge", 4): FragilityCurveSet._4ls_to_5ds,
            ("tsunami", "bridge", 4): FragilityCurveSet._4ls_to_5ds,
            ("tsunami", "pipeline", 4): FragilityCurveSet._4ls_to_5ds,
            ("tsunami", "road", 4): FragilityCurveSet._4ls_to_5ds,
            ("tsunami", "water_facility", 4): FragilityCurveSet._4ls_to_5ds,
            ("tsunami", "electric_facility", 4): FragilityCurveSet._4ls_to_5ds,
            ("hurricane", "bridge", 4): FragilityCurveSet._4ls_to_5ds,
        }

        if not (hazard_type, inventory_type, len(self.fragility_curves)) in ls_ds_dspatcher.keys():
            raise ValueError(inventory_type + " " + hazard_type + " damage analysis do not support " +
                             str(len(self.fragility_curves)) + " limit state")

        return ls_ds_dspatcher[(hazard_type, inventory_type, len(self.fragility_curves))](damage)

    def construct_expression_args_from_inventory(self, inventory_unit: dict):
        """

        Args:
            inventory_unit (dict): An inventory set.

        Returns:
            dict: Function parameters.

        """
        kwargs_dict = {}
        for parameters in self.curve_parameters:

            if parameters['name'] == "age_group" and ('age_group' not in inventory_unit['properties'] or
                                                      inventory_unit['properties']['age_group'] == ""):
                if 'year_built' in inventory_unit['properties'].keys() and inventory_unit['properties']['year_built'] \
                        is not None:
                    try:
                        yr_built = int(inventory_unit['properties']['year_built'])
                    except ValueError:
                        print("Non integer value found in year_built")
                        raise
                    age_group = 4  # for yr_built >= 2008
                    if yr_built < 1974:
                        age_group = 1
                    elif 1974 <= yr_built < 1987:
                        age_group = 2
                    elif 1987 <= yr_built < 1995:
                        age_group = 3
                    elif 1995 <= yr_built < 2008:
                        age_group = 4

                    kwargs_dict['age_group'] = age_group

            if parameters['name'] in inventory_unit['properties'] and \
                    inventory_unit['properties'][parameters['name']] is not None and \
                    inventory_unit['properties'][parameters['name']] != "":
                kwargs_dict[parameters['name']] = inventory_unit['properties'][parameters['name']]
        return kwargs_dict

    @staticmethod
    def _3ls_to_4ds(limit_states):
        """

        Args:
            limit_states (dict): Limit states.

        Returns:
            dict: Damage states.

        """
        limit_states = AnalysisUtil.float_dict_to_decimal(limit_states)
        damage_states = AnalysisUtil.float_dict_to_decimal({"DS_0": 0.0, "DS_1": 0.0, "DS_2": 0.0, "DS_3": 0.0})

        small_overlap = FragilityCurveSet.is_there_small_overlap(limit_states)

        if small_overlap:
            ds_overlap = FragilityCurveSet.adjust_for_small_overlap(small_overlap, limit_states, damage_states)

            damage_states['DS_0'] = ds_overlap[0]
            damage_states['DS_1'] = ds_overlap[1]
            damage_states['DS_2'] = ds_overlap[2]
            damage_states['DS_3'] = ds_overlap[3]

        else:
            damage_states['DS_0'] = 1 - limit_states["LS_0"]
            damage_states['DS_1'] = limit_states["LS_0"] - limit_states["LS_1"]
            damage_states['DS_2'] = limit_states["LS_1"] - limit_states["LS_2"]
            damage_states['DS_3'] = limit_states["LS_2"]

        return damage_states

    @staticmethod
    def _4ls_to_5ds(limit_states):
        """

        Args:
            limit_states (dict): Limit states.

        Returns:
            dict: Damage states.

        """
        limit_states = AnalysisUtil.float_dict_to_decimal(limit_states)
        damage_states = AnalysisUtil.float_dict_to_decimal({"DS_0": 0.0, "DS_1": 0.0, "DS_2": 0.0, "DS_3": 0.0,
                                                            "DS_4": 0.0})

        small_overlap = FragilityCurveSet.is_there_small_overlap(limit_states)

        if small_overlap:
            ds_overlap = FragilityCurveSet.adjust_for_small_overlap(small_overlap, limit_states, damage_states)

            damage_states['DS_0'] = ds_overlap[0]
            damage_states['DS_1'] = ds_overlap[1]
            damage_states['DS_2'] = ds_overlap[2]
            damage_states['DS_3'] = ds_overlap[3]
            damage_states['DS_4'] = ds_overlap[4]

        else:
            damage_states['DS_0'] = 1 - limit_states["LS_0"]
            damage_states['DS_1'] = limit_states["LS_0"] - limit_states["LS_1"]
            damage_states['DS_2'] = limit_states["LS_1"] - limit_states["LS_2"]
            damage_states['DS_3'] = limit_states["LS_2"] - limit_states["LS_3"]
            damage_states['DS_4'] = limit_states["LS_3"]

        return damage_states

    @staticmethod
    def _1ls_to_4ds(limit_states):
        """

        Args:
            limit_states (dict): Limit states.

        Returns:
            dict: Damage states.

        """
        limit_states = AnalysisUtil.float_dict_to_decimal(limit_states)
        damage_states = dict()
        damage_states['DS_0'] = 1 - limit_states["LS_0"]
        damage_states['DS_1'] = 0
        damage_states['DS_2'] = 0
        damage_states['DS_3'] = limit_states["LS_0"]

        return damage_states

    @staticmethod
    def _1ls_to_5ds(limit_states):
        """

        Args:
            limit_states (dict): Limit states.

        Returns:
            dict: Damage states.

        """
        limit_states = AnalysisUtil.float_dict_to_decimal(limit_states)
        damage_states = dict()
        damage_states['DS_0'] = 1 - limit_states["LS_0"]
        damage_states['DS_1'] = 0
        damage_states['DS_2'] = 0
        damage_states['DS_3'] = 0
        damage_states['DS_4'] = limit_states["LS_0"]

        return damage_states

    @staticmethod
    def is_there_small_overlap(limit_states):
        small_overlap = []
        keys = list(limit_states)
        for ls_index in range(len(limit_states)):
            for tmp_index in range(ls_index + 1, len(limit_states)):
                if limit_states[keys[ls_index]] < limit_states[keys[tmp_index]]:
                    # if previous limit state is less than the next, there's an overlap
                    small_overlap.append(ls_index)
                    break

        return small_overlap

    @staticmethod
    def adjust_for_small_overlap(small_overlap, limit_states, damage_states):
        """

        Args:
            small_overlap (obj): Overlap.
            limit_states (dict): Limit states.
            damage_states (dict): Damage states.

        Returns:
            list: Damage states overlap.

        """
        ls_overlap = list(limit_states.values())
        ds_overlap = [0.0] * len(damage_states)
        for index in range(len(damage_states)):
            ds_index = index
            # If the limit state is overlapped, find the next non-overlapping limit state
            if index in small_overlap:
                for tmp_index in range(index + 1, len(ls_overlap)):
                    if tmp_index not in small_overlap:
                        ds_index = tmp_index

            if index == 0:
                # Compute DS_0
                ds_overlap[index] = 1 - ls_overlap[ds_index]
            elif index == len(damage_states) - 1:
                # Compute last DS
                ds_overlap[index] = ls_overlap[index - 1]
            else:
                # If one of the limit state curves between the first and last are overlapped, they've been
                # eliminated and will be 0. If not, then compute the interval
                current_ls = index - 1
                if current_ls not in small_overlap:
                    ds_overlap[index] = ls_overlap[index - 1] - ls_overlap[ds_index]

        return ds_overlap

    @staticmethod
    def _initialize_limit_states(inventory_type):
        """

        Args:
            inventory_type (str): Inventory type..

        Returns:
            dict: Limit states..

        """
        if inventory_type == "building":
            output = {"LS_0": 0.0, "LS_1": 0.0, "LS_2": 0.0}
        elif inventory_type == "pipeline":
            output = {"LS_0": 0.0, "LS_1": 0.0, "LS_2": 0.0, "LS_3": 0.0}
        elif inventory_type == "bridge":
            output = {"LS_0": 0.0, "LS_1": 0.0, "LS_2": 0.0, "LS_3": 0.0}
        elif inventory_type == "electric_facility":
            output = {"LS_0": 0.0, "LS_1": 0.0, "LS_2": 0.0, "LS_3": 0.0}
        elif inventory_type == "road":
            output = {"LS_0": 0.0, "LS_1": 0.0, "LS_2": 0.0, "LS_3": 0.0}
        elif inventory_type == "water_facility":
            output = {"LS_0": 0.0, "LS_1": 0.0, "LS_2": 0.0, "LS_3": 0.0}
        elif inventory_type == "Transmission Towers":
            output = {"LS_0": 0.0}
        elif inventory_type == "Distribution Poles":
            output = {"LS_0": 0.0}
        else:
            output = {"LS_0": 0.0, "LS_1": 0.0, "LS_2": 0.0}

        return output



from pyincore.analyses.waterfacilityrestoration import WaterFacilityRestoration


class WaterFacilityRestorationUtil:
    def __init__(self, inventory_restoration_map, pf_results, time_results, time_interval, pf_interval, end_time):

        # merge inventory_restoration_map with pf and timetables
        inventory_restoration_map_df = inventory_restoration_map.get_dataframe_from_csv()
        pf_results_df = pf_results.get_dataframe_from_csv()
        time_results_df = time_results.get_dataframe_from_csv()
        self.pf_results_df = inventory_restoration_map_df.merge(pf_results_df, on="restoration_id").set_index('guid')
        self.time_results_df = inventory_restoration_map_df.merge(time_results_df, on="restoration_id").set_index(
            'guid')

        self.time_interval = time_interval
        self.pf_interval = pf_interval
        self.end_time = end_time

    def get_restoration_time(self, guid, damage_state="DS_0", pf=0.99):
        if pf > 1:
            raise ValueError("Percentage of functionality should not be larger than 1!")

        state = "time_" + damage_state.replace("DS", "PF")
        df = self.pf_results_df.loc[guid].reset_index(drop=True)
        # round up and get the closest
        time = df.loc[(df["percentage_of_functionality"] >= pf) & (df["percentage_of_functionality"] <
                                                                   pf+self.pf_interval), state].values[0]

        return time

    def get_percentage_func(self, guid, damage_state="DS_0", time=1):
        if time > self.end_time:
            raise ValueError("restore time should not be larger than end time for restoration model!")

        state = damage_state.replace("DS", "PF")
        df = self.time_results_df.loc[guid].reset_index(drop=True)
        # round up and get the closest
        pf = df.loc[(df["time"] >= time) & df['time'] < time+self.time_interval, state].values[0]

        return pf



# Copyright (c) 2022 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

# TODO: exception handling for validation and set methods
from pyincore import DataService, AnalysisUtil, Earthquake, Tornado, Tsunami, Hurricane, Flood
from pyincore.dataset import Dataset
import typing


class BaseAnalysis:
    """Superclass that defines the specification for an IN-CORE analysis.
    Implementations of BaseAnalysis should implement get_spec and return their own
    specifications.

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        self.spec = self.get_spec()
        self.client = incore_client
        self.data_service = DataService(self.client)

        # initialize parameters, input_datasets, output_datasets, etc
        self.parameters = {}
        for param in self.spec['input_parameters']:
            self.parameters[param['id']] = {'spec': param, 'value': None}

        self.input_datasets = {}
        for input_dataset in self.spec['input_datasets']:
            self.input_datasets[input_dataset['id']] = {'spec': input_dataset, 'value': None}

        self.input_hazards = {}
        if 'input_hazards' in self.spec:
            for input_hazards in self.spec['input_hazards']:
                self.input_hazards[input_hazards['id']] = {'spec': input_hazards, 'value': None}

        self.output_datasets = {}
        for output_dataset in self.spec['output_datasets']:
            self.output_datasets[output_dataset['id']] = {'spec': output_dataset, 'value': None}

    def get_spec(self):
        """Get basic specifications.

        Note:
            The get_spec will be called exactly once per instance (during __init__),
            so children should not assume that they can do weird dynamic magic during this call.
            See the example spec at the bottom of this file.

        Returns:
            obj: A JSON object of basic specifications of the analysis.

        """
        return {
            'name': 'base-analysis',
            'description': 'this should be replaced by analysis spec',
            'input_parameters': [
            ],
            'input_datasets': [
            ],
            'output_datasets': [
            ]
        }

    def load_remote_input_dataset(self, analysis_param_id, remote_id):
        """Convenience function for loading a remote dataset by id.

        Args:
            analysis_param_id (str): ID of the input Dataset in the specifications.
            remote_id (str):  ID of the Dataset in the Data service.

        """
        dataset = Dataset.from_data_service(remote_id, self.data_service)

        # TODO: Need to handle failing to set input dataset.
        self.set_input_dataset(analysis_param_id, dataset)

    def get_name(self):
        """Get the analysis name."""
        return self.spec['name']

    def get_description(self):
        """Get the description of an analysis."""
        return self.spec['description']

    def get_parameters(self):
        """Get the dictionary of analysis' parameters."""
        param = {}
        for key in self.parameters.keys():
            param[key] = self.parameters[key]['value']
        return param

    def get_parameter(self, par_id):
        """Get or set the analysis parameter value. Setting a parameter to a new value
        will return True or False on error."""
        return self.parameters[par_id]['value']

    def set_parameter(self, par_id, parameter):
        result = self.validate_parameter(self.parameters[par_id]['spec'], parameter)

        if result[0]:
            self.parameters[par_id]['value'] = parameter
            return True
        else:
            print("Error setting parameter: " + result[1])
            return False

    def get_input_datasets(self):
        """Get the dictionary of the input datasets of an analysis."""
        inputs = {}
        for key in self.input_datasets.keys():
            inputs[key] = self.input_datasets[key]['value']
        return inputs

    def get_input_dataset(self, ds_id):
        """Get or set the analysis dataset. Setting the dataset to a new value
        will return True or False on error."""
        return self.input_datasets[ds_id]['value']

    def set_input_dataset(self, ds_id, dataset):
        result = self.validate_input_dataset(self.input_datasets[ds_id]['spec'], dataset)
        if result[0]:
            self.input_datasets[ds_id]['value'] = dataset
            return True
        else:
            print(result[1])
            return False

    def get_input_hazards(self):
        """Get the dictionary of the input hazards of an analysis."""
        inputs = {}
        for key in self.input_hazards.keys():
            inputs[key] = self.input_hazards[key]['value']
        return inputs

    def get_input_hazard(self, hz_id):
        """Get or set the analysis dataset. Setting the hazard to a new value
        will return True or False on error."""
        return self.input_hazards[hz_id]['value']

    def set_input_hazard(self, hz_id, hazard):
        result = self.validate_input_hazard(self.input_hazards[hz_id]['spec'], hazard)
        if result[0]:
            self.input_hazards[hz_id]['value'] = hazard
            return True
        else:
            print(result[1])
            return False

    def create_hazard_object_from_input_params(self):
        """Create hazard object from input parameters."""
        hazard_object = self.get_input_hazard("hazard")
        hazard_type = self.get_parameter("hazard_type")
        hazard_dataset_id = self.get_parameter("hazard_id")

        # either hazard object or hazard id + hazard type must be provided
        if hazard_object is None and (hazard_type is None or hazard_dataset_id is None):
            raise ValueError("Either hazard object or hazard id + hazard type must be provided")

        # create hazard object from remote
        elif hazard_object is None and hazard_type is not None and hazard_dataset_id is not None:
            hazard_object = BaseAnalysis._create_hazard_object(hazard_type, hazard_dataset_id, self.hazardsvc)

        # use hazard object
        else:
            hazard_type = hazard_object.hazard_type
            hazard_dataset_id = hazard_object.id

        return hazard_object, hazard_type, hazard_dataset_id

    @staticmethod
    def _create_hazard_object(hazard_type, hazard_dataset_id, hazardsvc):
        """Helper function to create hazard object from hazard type and hazard dataset id."""
        if hazard_type == "earthquake":
            hazard_object = Earthquake.from_hazard_service(hazard_dataset_id, hazardsvc)
        elif hazard_type == "tornado":
            hazard_object = Tornado.from_hazard_service(hazard_dataset_id, hazardsvc)
        elif hazard_type == "tsunami":
            hazard_object = Tsunami.from_hazard_service(hazard_dataset_id, hazardsvc)
        elif hazard_type == "hurricane":
            hazard_object = Hurricane.from_hazard_service(hazard_dataset_id, hazardsvc)
        elif hazard_type == "flood":
            hazard_object = Flood.from_hazard_service(hazard_dataset_id, hazardsvc)
        else:
            raise ValueError("The provided hazard type is not supported.")

        return hazard_object

    def get_output_datasets(self):
        """Get the output dataset of the analysis."""
        outputs = {}
        for key in self.output_datasets.keys():
            outputs[key] = self.output_datasets[key]['value']
        return outputs

    def get_output_dataset(self, ds_id):
        """Get or set the output dataset. Setting the output dataset to a new value
        will return True or False on error."""
        return self.output_datasets[ds_id]['value']

    def set_output_dataset(self, ds_id, dataset):
        if self.validate_output_dataset(self.output_datasets[ds_id]['spec'], dataset)[0]:
            self.output_datasets[ds_id]['value'] = dataset
            return True
        else:
            # TODO handle error message
            return False

    @staticmethod
    def validate_parameter_nested(parameter, parameter_spec):
        is_valid = True
        err_msg = ''

        if type(parameter_spec['type']) is typing._GenericAlias:
            if not (type(parameter) is parameter_spec['type'].__origin__):
                is_valid = False
                err_msg = 'container parameter type does not match - spec: ' + str(parameter_spec)
            elif not (all(isinstance(s, parameter_spec['type'].__args__[0]) for s in parameter)):
                is_valid = False
                err_msg = 'element parameter type does not match - spec: ' + str(parameter_spec)
        else:
            if not type(parameter) is parameter_spec['type']:
                is_valid = False
                err_msg = 'parameter type does not match - spec: ' + str(parameter_spec)

        return is_valid, err_msg

    def validate_parameter(self, parameter_spec, parameter):
        """Match parameter by type.

        Args:
            parameter_spec (obj): Specifications of parameters.
            parameter (obj): Parameter description.

        Returns:
            bool, str: Parameter validity, True if valid, False otherwise. Error message.

        """
        is_valid = True
        err_msg = ''

        if parameter_spec['required']:
            if parameter is None:
                is_valid = False
                err_msg = 'required parameter is missing - spec: ' + str(parameter_spec)
            else:
                is_valid, err_msg = self.validate_parameter_nested(parameter, parameter_spec)
        else:
            if parameter is not None:
                is_valid, err_msg = self.validate_parameter_nested(parameter, parameter_spec)

        return is_valid, err_msg

    @staticmethod
    def validate_input_dataset(dataset_spec, dataset):
        """Match input dataset by type.

        Args:
            dataset_spec (obj): Specifications of datasets.
            dataset (obj): Dataset description.

        Returns:
            bool, str: Dataset validity, True if valid, False otherwise. Error message.

        """
        is_valid = True
        err_msg = ''

        if not isinstance(dataset, type(None)):
            # if dataset is not none, check data type
            if not (dataset.data_type in dataset_spec['type']):
                # if dataset type is not equal to spec, then return false
                is_valid = False
                err_msg = 'dataset type does not match - ' + 'given type: ' + \
                          dataset.data_type + ' spec types: ' + str(dataset_spec['type'])
        else:
            # if dataset is none, check 'requirement'
            if dataset_spec['required']:
                # if dataset is 'required', return false
                is_valid = False
                err_msg = 'required dataset is missing - spec: ' + str(dataset_spec)
        return is_valid, err_msg

    @staticmethod
    def validate_input_hazard(hazard_spec, hazard):
        """Validate input hazard.

        Args:
            hazard_spec (obj): Specifications of hazard.
            hazard (obj): Hazard description.

        Returns:
            bool, str: Hazard validity, True if valid, False otherwise. Error message.

        """
        is_valid = True
        err_msg = ''

        if not isinstance(hazard, type(None)):
            # if hazard is not none, check hazard instance type
            is_valid = False
            for hazard_type in hazard_spec['type']:
                if hazard.hazard_type == hazard_type:
                    is_valid = True
                    break
            if not is_valid:
                err_msg = 'hazard type does not match - ' + 'given type: ' + \
                          hazard.hazard_type + ' spec types: ' + str(hazard_spec['type'])
        else:
            # if hazard is none, check 'requirement'
            if hazard_spec['required']:
                is_valid = False
                err_msg = 'required hazard is missing - spec: ' + str(hazard_spec)
        return is_valid, err_msg

    @staticmethod
    def validate_output_dataset(dataset_spec, dataset):
        """Match output dataset by type.

        Args:
            dataset_spec (obj): Specifications of datasets.
            dataset (obj): Dataset description.

        Returns:
            bool, str: Dataset validity, True if valid, False otherwise. Error message.

        """
        is_valid = True
        err_msg = ''
        if not (dataset.data_type is dataset_spec['type']):
            is_valid = False
            err_msg = 'dataset type does not match'
        return is_valid, err_msg

    """ convenience function(s) for setting result data as a csv """

    def set_result_csv_data(self, result_id, result_data, name, source='file', index=False):
        if name is None:
            name = self.spec["name"] + "-result"

        if not name.endswith(".csv"):
            name = name + ".csv"

        dataset_type = self.output_datasets[result_id]["spec"]["type"]
        dataset = None

        if source == 'file':
            dataset = Dataset.from_csv_data(result_data, name, dataset_type)
        elif source == 'dataframe':
            dataset = Dataset.from_dataframe(result_data, name, dataset_type, index)

        self.set_output_dataset(result_id, dataset)

    def set_result_json_data(self, result_id, result_data, name, source='file'):
        if name is None:
            name = self.spec["name"] + "-result"

        if not name.endswith(".json"):
            name = name + ".json"

        dataset_type = self.output_datasets[result_id]["spec"]["type"]
        dataset = None
        if source == 'file':
            dataset = Dataset.from_json_data(result_data, name, dataset_type)

        self.set_output_dataset(result_id, dataset)

    def run_analysis(self):
        """ Validates and runs the analysis."""
        for dataset_spec in self.spec['input_datasets']:
            ds_id = dataset_spec["id"]
            result = self.validate_input_dataset(dataset_spec, self.input_datasets[ds_id]["value"])

            if not result[0]:
                print("Error reading dataset: " + result[1])
                return result

        # TODO: We will iteratively roll out input hazard; once it's done, we will remove this if block
        if 'input_hazards' in self.spec:
            for hazard_spec in self.spec['input_hazards']:
                hz_id = hazard_spec["id"]
                result = self.validate_input_hazard(hazard_spec, self.input_hazards[hz_id]["value"])

                if not result[0]:
                    print("Error reading hazard: " + result[1])
                    return result

        for parameter_spec in self.spec['input_parameters']:
            par_id = parameter_spec["id"]
            result = self.validate_parameter(parameter_spec, self.get_parameter(par_id))

            if not result[0]:
                print("Error reading parameter: " + result[1])
                return result

        return self.run()

    def run(self):
        return True

    def show_gdocstr_docs(self):
        return AnalysisUtil.create_gdocstr_from_spec(self.get_spec())



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import urllib

from pyincore import IncoreClient
from pyincore.dfr3service import Dfr3Service


class RepairService(Dfr3Service):
    """Fragility service client.

    Args:
        client (IncoreClient): Service authentication.

    """

    def __init__(self, client: IncoreClient):
        self.client = client
        self.base_dfr3_url = urllib.parse.urljoin(client.service_url,
                                                  'dfr3/api/repairs/')

        super(RepairService, self).__init__(client)

    def get_dfr3_sets(self, hazard_type: str = None, inventory_type: str = None,
                      author: str = None, creator: str = None, space: str = None,
                      skip: int = None, limit: int = None, timeout=(30, 600), **kwargs):
        """Get the set of repair data, curves.

        Args:
            hazard_type (str): Hazard type filter, default None.
            inventory_type (str): Inventory type, default None.
            author (str): Repair set creator’s username, default None.
            creator (str): Repair creator’s username, default None.
            space (str): Name of space, default None.
            skip (int):  Skip the first n results, default None.
            limit (int): Limit number of results to return, default None.
            timeout (tuple): Timeout for the request, default (30, 600).
            **kwargs: Additional arguments.

        Returns:
            obj: HTTP response with search results.

        """
        url = self.base_dfr3_url
        payload = {}

        if hazard_type is not None:
            payload['hazard'] = hazard_type
        if inventory_type is not None:
            payload['inventory'] = inventory_type
        if author is not None:
            payload['author'] = author
        if creator is not None:
            payload['creator'] = creator
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit
        if space is not None:
            payload['space'] = space

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)
        return r.json()



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import string


class INDPComponents:
    """
    This class stores components of a network

    Attributes
    ----------
    components : list
        List of components of the network
    num_components : int
        Number of components of the network
    gc_size : int
        Size of the largest component of the network
    """

    def __init__(self):
        self.components = []
        self.num_components = 0
        self.gc_size = 0

    def add_component(self, members, excess_supply):
        """
        This function adds a components

        Parameters
        ----------
        members : list
            List of nodes in the component
        excess_supply : float
            Excess supply within the component

        Returns
        -------
        None.

        """
        self.components.append((members, excess_supply))

    def to_csv_string(self):
        """
        Convert the list of components to a string

        Returns
        -------
            : str
            List of components as a string

        """
        comp_strings = []
        for c in self.components:
            comp = c[0]
            supp = c[1]
            comp_string = "/".join(c[0])
            comp_string += ":" + str(c[1])
            comp_strings.append(comp_string)
        return ",".join(comp_strings)

    @classmethod
    def calculate_components(clss, m, net, t=0, layers=None):
        """
        Find the components and the corresponding excess supply

        Parameters
        ----------
        m : gurobi.Model
            The object containing the solved optimization problem.
        net : networkx.DiGraph
            The networkx graph object that stores node, arc, and interdependency information
        t : int
            Time step. The default is zero.
        layers : list
            List of layers in the analysis

        Returns
        -------
        indp_components : :class:`~INDPComponents`
            The object containing the components

        """
        if layers is None:
            layers = [1, 2, 3]
        indp_components = INDPComponents()
        components = net.get_clusters(layers[0])
        indp_components.num_components = len(components)
        indp_components.gc_size = net.gc_size(layers[0])
        for c in components:
            total_excess_supply = 0.0
            members = []
            for n in c:
                members.append(str(n[0]) + "." + str(n[1]))
                excess_supply = 0.0
                excess_supply += m.getVarByName('delta+_' + str(n) + "," + str(t)).x
                excess_supply += -m.getVarByName('delta-_' + str(n) + "," + str(t)).x
                total_excess_supply += excess_supply
            indp_components.add_component(members, total_excess_supply)
        return indp_components

    @classmethod
    def from_csv_string(clss, csv_string):
        """
        This functions reads the components data from a string as a :class:`~INDPComponents` object

        Parameters
        ----------
        csv_string : str
            The string containing the component data

        Returns
        -------
        indp_components : :class:`~INDPComponents`
            The object containing the components

        """
        indp_components = INDPComponents()
        comps = csv_string
        for comp in comps:
            data = string.split(comp, ":")
            members = string.split(data[0], "/")
            supp = str.strip(data[1])
            indp_components.add_component(members, float(supp))
        return indp_components



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


class Units:

    deg_to_rad = 0.0174533
    rad_to_deg = 57.2958

    ms_to_msec = 1
    msec_to_ms = 1
    ms_to_ins = 39.3701
    msec_to_ins = 39.3701
    ms_to_fts = 3.28084
    msec_to_fts = 3.28084
    ms_to_mph = 2.23694
    mps_to_mph = 2.23694
    msec_to_mph = 2.23694

    ins_to_msec = 0.0254
    ins_to_ms = 0.0254
    ins_to_fts = 0.0833333
    ins_to_mph = 0.0568182

    fts_to_msec = 0.3048
    fts_to_ms = 0.3048
    fts_to_ins = 12
    fts_to_mph = 0.681818

    mph_to_ins = 17.6
    mph_to_fts = 1.46667
    mph_to_ms = 0.44704
    mph_to_mps = 0.44704
    mph_to_msec = 0.44704

    m_to_cm = 100.0
    m_to_in = 39.3701
    m_to_ft = 3.28084

    in_to_cm = 2.54
    in_to_m = 0.0254
    in_to_ft = 0.0833333

    cm_to_m = 0.01
    cm_to_in = 0.3937
    cm_to_ft = 0.0328

    ft_to_cm = 30.48
    ft_to_in = 12
    ft_to_m = 0.3048

    hr_to_min = 60
    min_to_hr = 0.01666
    min_to_s = 60
    s_to_min = 0.01666
    hr_to_s = 3600
    s_to_hr = 0.00028

    @staticmethod
    def convert_hazard(hazard_value, original_demand_units, requested_demand_units):
        converted_hazard_value = hazard_value
        if original_demand_units.lower() != requested_demand_units.lower():
            conversion = f"{original_demand_units.lower().replace('/', '')}_to_" \
                         f"{requested_demand_units.lower().replace('/', '')}"
            try:
                conversion_value = getattr(Units, conversion)
                converted_hazard_value = conversion_value * hazard_value
            except AttributeError:
                raise ValueError(f"We don't support the conversion from {original_demand_units} "
                                 f"to {requested_demand_units}")
        else:
            return converted_hazard_value

        return converted_hazard_value



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

from setuptools import setup, find_packages

# version number of pyincore
version = '1.15.0'

with open("README.rst", encoding="utf-8") as f:
    readme = f.read()

setup(
    name='pyincore',
    version=version,
    description='IN-CORE analysis tool python package',
    long_description=readme,
    long_description_content_type='text/x-rst',

    url='https://incore.ncsa.illinois.edu',

    license="Mozilla Public License v2.0",

    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
    ],

    keywords=[
        "infrastructure",
        "resilience",
        "hazards",
        "data discovery",
        "IN-CORE",
        "earthquake",
        "tsunami",
        "tornado",
        "hurricane",
        "dislocation"
    ],

    packages=find_packages(where=".", exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    package_data={
        '': ['*.ini']
    },

    python_requires=">=3.9",

    install_requires=[
        'fiona>=1.9.5',
        'geopandas>=0.14.0',
        'matplotlib>=3.8.0',
        'networkx>=3.2.1',
        'numpy>=1.26.0,<2.0a0',
        'pandas>=2.1.2',
        'pyomo>=6.6.2',
        'pyproj>=3.6.1',
        'rasterio>=1.3.9',
        'rtree>=1.1.0',
        'scipy>=1.11.3',
        'shapely>=2.0.2',
    ],

    extras_require={
        'test': [
            'pycodestyle>=2.6.0',
            'pytest>=3.9.0',
            'python-jose>=3.0',
        ]
    },

    project_urls={
        'Bug Reports': 'https://github.com/IN-CORE/pyincore/issues',
        'Source': 'https://github.com/IN-CORE/pyincore',
    },
)



# Copyright (c) 2022 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import copy
import numpy as np
import pandas as pd
import networkx as nx

from typing import List
from pyincore import BaseAnalysis, NetworkDataset, NetworkUtil
from pyincore.analyses.wfnfunctionality.wfnfunctionalityutil import WfnFunctionalityUtil


class WfnFunctionality(BaseAnalysis):
    """Computes water facility network functionality.

    Args:
        incore_client: Service client with authentication info
    """

    def __init__(self, incore_client):
        super(WfnFunctionality, self).__init__(incore_client)

    def run(self):
        """Execute water facility network functionality analysis """

        # Obtain tank nodes
        tank_nodes = self.get_parameter('tank_node_list')

        # Obtain pump station nodes
        pumpstation_nodes = self.get_parameter('pumpstation_node_list')

        # Get network dataset
        network_dataset = NetworkDataset.from_dataset(self.get_input_dataset('wfn_network'))
        edges_wfl_gdf = network_dataset.links.get_dataframe_from_shapefile()

        nodes_wfn_gdf = network_dataset.nodes.get_dataframe_from_shapefile()
        edges_wfl_gdf['weight'] = edges_wfl_gdf.loc[:, 'length']

        G_wfn = network_dataset.get_graph_networkx()

        # network test
        fromnode_fld_name = 'fromnode'
        tonode_fld_name = 'tonode'
        nodenwid_fld_name = 'nodenwid'

        node_id_validation = NetworkUtil.validate_network_node_ids(
            network_dataset, fromnode_fld_name, tonode_fld_name, nodenwid_fld_name)

        if node_id_validation is False:
            print("ID in from or to node field doesn't exist in the node dataset")
            return False

        # Get water facility damage states
        wf_dmg_fs = self.get_input_dataset('wf_sample_failure_state').get_dataframe_from_csv()
        wf_sample_df = pd.DataFrame(
            np.array([np.array(wf_dmg_fs.failure.values[i].split(',')).astype('int')
                      for i in np.arange(wf_dmg_fs.shape[0])]),
            index=wf_dmg_fs.guid.values)

        # Get pipeline damage states
        pp_dmg_fs = self.get_input_dataset('pp_sample_failure_state').get_dataframe_from_csv()
        pp_sample_df = pd.DataFrame(
            np.array([np.array(pp_dmg_fs.failure.values[i].split(',')).astype('int')
                      for i in np.arange(pp_dmg_fs.shape[0])]),
            index=pp_dmg_fs.guid.values)

        # Get the sample number
        num_samples = wf_sample_df.shape[1]
        sampcols = ['s' + samp for samp in np.arange(num_samples).astype(str)]

        # Compose the corresponding dataframes based on columns for water facilities and pipelines
        wf_sample_df.columns = sampcols
        wf_sample_df1 = nodes_wfn_gdf.loc[:, ['guid', 'nodenwid']].set_index('guid').join(wf_sample_df)
        wf_sample_df1 = wf_sample_df1.fillna(1)

        pp_sample_df.columns = sampcols
        pp_sample_df1 = edges_wfl_gdf.loc[:, ['guid', 'fromnode', 'tonode']].set_index('guid').join(pp_sample_df)

        # Obtain distribution nodes based on user input
        distribution_nodes = list(set(list(G_wfn.nodes)) - set(tank_nodes) - set(pumpstation_nodes))

        (fs_results, fp_results) = self.wfn_functionality(distribution_nodes, pumpstation_nodes, num_samples,
                                                          sampcols, wf_sample_df1, pp_sample_df1, G_wfn)

        self.set_result_csv_data("sample_failure_state",
                                 fs_results, name=self.get_parameter("result_name") + "_failure_state",
                                 source="dataframe")
        self.set_result_csv_data("failure_probability",
                                 fp_results,
                                 name=self.get_parameter("result_name") + "_failure_probability",
                                 source="dataframe")

        return True

    def wfn_functionality(self, distribution_nodes, pumpstation_nodes, num_samples, sampcols, wf_sample_df1,
                          pp_sample_df1, G_wfn):
        """
        Run Water facility network functionality analysis.

        Args:
            distribution_nodes (list): distribution nodes
            pumpstation_nodes (list): pump station nodes
            num_samples (int): number of simulations
            sampcols (list): list of number samples. e.g. "s0, s1,..."
            wf_sample_df1 (dataframe): water facility mcs failure sample dataframe
            pp_sample_df1 (dataframe): pipeline mcs failure sample dataframe
            G_wfn (networkx object): constructed network

        Returns:
            fs_results (list): A list of dictionary with id/guid and failure state for N samples
            fp_results (list): A list dictionary with failure probability and other data/metadata.

        """

        # a distance of M denotes disconnection
        M = 9999

        func_wf_df = pd.DataFrame(np.zeros((len(distribution_nodes), num_samples)), index=distribution_nodes,
                                  columns=sampcols)

        for si, scol in enumerate(sampcols):
            nodestate_wfn = wf_sample_df1.loc[:, ['nodenwid', scol]]
            linkstate_wfn = pp_sample_df1.loc[:, ['fromnode', 'tonode', scol]]
            badlinks_wfn = WfnFunctionalityUtil.get_bad_edges(G_wfn, nodestate_wfn, linkstate_wfn, scol)
            badlinkdict_wfn = {k: {'weight': M} for k in badlinks_wfn}
            G1_wfn = copy.deepcopy(G_wfn)
            nx.set_edge_attributes(G1_wfn, badlinkdict_wfn)
            res_ep = WfnFunctionalityUtil.network_shortest_paths(G1_wfn, pumpstation_nodes, distribution_nodes)
            func_wf_df.loc[distribution_nodes, scol] = (res_ep < M) * 1

        # Use nodenwid index to get its guid
        func_wf_df.index = func_wf_df.index.map(np.int64)

        fs_temp = pd.merge(func_wf_df, wf_sample_df1["nodenwid"], left_index=True, right_on="nodenwid",
                           how='left').drop(columns=["nodenwid"])
        fp_temp = fs_temp.copy(deep=True)

        # shape the dataframe into failure probability and failure samples
        fs_temp['failure'] = fs_temp.astype(str).apply(','.join, axis=1)
        fs_results = fs_temp.filter(['failure'])
        fs_results.reset_index(inplace=True)
        fs_results = fs_results.rename(columns={'index': 'guid'})

        # calculate failure probability
        # count of 0s divided by sample size
        fp_temp["failure_probability"] = (num_samples - fp_temp.sum(axis=1).astype(int)) / num_samples
        fp_results = fp_temp.filter(['failure_probability'])
        fp_results.reset_index(inplace=True)
        fp_results = fp_results.rename(columns={'index': 'guid'})

        return fs_results, fp_results

    def get_spec(self):
        """Get specifications of the water facility network functionality analysis.
        Returns:
            obj: A JSON object of specifications of the WFN functionality analysis.
        """
        return {
            'name': 'wfn-functionality',
            'description': 'water facility network functionality analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'tank_node_list',
                    'required': True,
                    'description': "list of tank nodes",
                    'type': List[int]
                },
                {
                    'id': 'pumpstation_node_list',
                    'required': True,
                    'description': "list of pump station nodes",
                    'type': List[int]
                }
            ],
            'input_datasets': [
                {
                    'id': 'wfn_network',
                    'required': True,
                    'description': 'Water Facility Network Dataset',
                    'type': ['incore:waterNetwork'],
                },
                {
                    'id': 'wf_sample_failure_state',
                    'required': True,
                    'description': 'CSV file of failure state for each sample. Output from MCS analysis',
                    'type': ['incore:sampleFailureState']
                },
                {
                    'id': 'pp_sample_failure_state',
                    'required': True,
                    'description': 'CSV file of failure state for each sample. Output from MCS analysis',
                    'type': ['incore:sampleFailureState']
                }
            ],
            'output_datasets': [
                {
                    'id': 'failure_probability',
                    'description': 'CSV file of failure probability',
                    'type': 'incore:failureProbability'
                },
                {
                    'id': 'sample_failure_state',
                    'description': 'CSV file of failure state for each sample',
                    'type': 'incore:sampleFailureState'
                },
            ]
        }
